import numpy as np
import pandas as pd
import statsmodels.api as sm
from statsmodels.tsa.stattools import coint

from qubx.utils import sbp


def percentile_rank(x: np.ndarray, v, pctls=np.arange(1, 101)):
    """
    Find percentile rank of value v
    :param x: values array
    :param v: vakue to be ranked
    :param pctls: percentiles
    :return: rank

    >>> percentile_rank(np.random.randn(1000), 1.69)
    >>> 95
    >>> percentile_rank(np.random.randn(1000), 1.69, [10,50,100])
    >>> 2
    """
    return np.argmax(np.sign(np.append(np.percentile(x, pctls), np.inf) - v))


def compare_to_norm(xs, xranges=None):
    """
    Compare distribution from xs against normal using estimated mean and std
    """
    import scipy.stats as stats
    import matplotlib.pyplot as plt
    import seaborn as sns

    _m, _s = np.mean(xs), np.std(xs)
    fit = stats.norm.pdf(sorted(xs), _m, _s)

    sbp(12, 1)
    plt.plot(sorted(xs), fit, "r--", lw=2, label="N(%.2f, %.2f)" % (_m, _s))
    plt.legend(loc="upper right")

    sns.kdeplot(xs, color="g", label="Data", fill=True)
    if xranges is not None and len(xranges) > 1:
        plt.xlim(xranges)
    plt.legend(loc="upper right")

    sbp(12, 2)
    stats.probplot(xs, dist="norm", sparams=(_m, _s), plot=plt)


def kde(array, cut_down=True, bw_method="scott"):
    """
    Kernel dense estimation
    """
    from scipy.stats import gaussian_kde

    if cut_down:
        bins, counts = np.unique(array, return_counts=True)
        f_mean = counts.mean()
        f_above_mean = bins[counts > f_mean]
        if len(f_above_mean) > 0:
            bounds = [f_above_mean.min(), f_above_mean.max()]
            array = array[np.bitwise_and(bounds[0] < array, array < bounds[1])]

    return gaussian_kde(array, bw_method=bw_method)


def hurst(series: np.array, max_lag: int = 20):
    """
    Simplest Hurst exponent helps test whether the time series is:
    (1) A Random Walk (H ~ 0.5)
    (2) Trending (H > 0.5)
    (3) Mean reverting (H < 0.5)
    """
    tau, lagvec = [], []

    # Step through the different lags
    for lag in range(2, max_lag):
        # Produce price different with lag
        pp = np.subtract(series[lag:], series[:-lag])

        # Write the different lags into a vector
        lagvec.append(lag)

        # Calculate the variance of the difference
        tau.append(np.sqrt(np.std(pp)))

    # Linear fit to a double-log graph to get power
    m = np.polyfit(np.log10(lagvec), np.log10(tau), 1)

    # Calculate hurst
    return m[0] * 2


def half_life(price: pd.Series) -> int:
    """
    Half-life is the period of time it takes for the price to revert back to the mean.
    """
    xs_lag = price.shift(1).bfill()
    xs_ret = price.diff().bfill()
    res = sm.OLS(xs_ret, sm.add_constant(xs_lag)).fit()
    return int(-np.log(2) / res.params.iloc[1])


def cointegration_test(p1: pd.Series, p2: pd.Series, alpha: float = 0.05) -> tuple[bool, float]:
    p1, p2 = p1.dropna().align(p2.dropna(), join="inner")
    _, pvalue, _ = coint(p1, p2)
    return bool(pvalue < alpha), float(pvalue)



import re, os
from typing import Dict, List, Set, Union, Optional, Iterator, Iterable, Any
from os.path import exists, join
import numpy as np
import pandas as pd
import pyarrow as pa
from pyarrow import csv
import psycopg as pg
from functools import wraps

from qubx import logger
from qubx.core.series import TimeSeries, OHLCV, time_as_nsec, Quote, Trade
from qubx.pandaz.utils import ohlc_resample, srows
from qubx.utils.time import infer_series_frequency, handle_start_stop
from psycopg.types.datetime import TimestampLoader

_DT = lambda x: pd.Timedelta(x).to_numpy().item()
D1, H1 = _DT("1D"), _DT("1h")
MS1 = 1_000_000

DEFAULT_DAILY_SESSION = (_DT("00:00:00.100"), _DT("23:59:59.900"))
STOCK_DAILY_SESSION = (_DT("9:30:00.100"), _DT("15:59:59.900"))
CME_FUTURES_DAILY_SESSION = (_DT("8:30:00.100"), _DT("15:14:59.900"))


class NpTimestampLoader(TimestampLoader):
    def load(self, data) -> np.datetime64:
        dt = super().load(data)
        return np.datetime64(dt)


def _recognize_t(t: Union[int, str], defaultvalue, timeunit) -> int:
    if isinstance(t, (str, pd.Timestamp)):
        try:
            return np.datetime64(t, timeunit)
        except:
            pass
    return defaultvalue


def _time(t, timestamp_units: str) -> int:
    t = int(t) if isinstance(t, float) else t
    if timestamp_units == "ns":
        return np.datetime64(t, "ns").item()
    return np.datetime64(t, timestamp_units).astype("datetime64[ns]").item()


def _find_column_index_in_list(xs, *args):
    xs = [x.lower() for x in xs]
    for a in args:
        ai = a.lower()
        if ai in xs:
            return xs.index(ai)
    raise IndexError(f"Can't find any from {args} in list: {xs}")


_FIND_TIME_COL_IDX = lambda column_names: _find_column_index_in_list(
    column_names, "time", "timestamp", "datetime", "date", "open_time", "ts"
)


class DataTransformer:

    def __init__(self) -> None:
        self.buffer = []
        self._column_names = []

    def start_transform(
        self,
        name: str,
        column_names: List[str],
        start: str | None = None,
        stop: str | None = None,
    ):
        self._column_names = column_names
        self.buffer = []

    def process_data(self, rows_data: Iterable) -> Any:
        if rows_data is not None:
            self.buffer.extend(rows_data)

    def collect(self) -> Any:
        return self.buffer


class DataReader:

    def get_names(self, **kwargs) -> List[str]:
        raise NotImplemented()

    def read(
        self,
        data_id: str,
        start: str | None = None,
        stop: str | None = None,
        transform: DataTransformer = DataTransformer(),
        chunksize=0,
        **kwargs,
    ) -> Iterator | List:
        raise NotImplemented()

    def get_aux_data_ids(self) -> Set[str]:
        """
        Returns list of all auxiliary data IDs available for this data reader
        """

        def _list_methods(cls):
            _meth = []
            for k, s in cls.__dict__.items():
                if k.startswith("get_") and k not in ["get_names", "get_aux_data_ids", "get_aux_data"] and callable(s):
                    _meth.append(k[4:])
            return _meth

        _d_ids = _list_methods(self.__class__)
        for bc in self.__class__.__bases__:
            _d_ids.extend(_list_methods(bc))
        return set(_d_ids)

    def get_aux_data(self, data_id: str, **kwargs) -> Any:
        """
        Returns auxiliary data for the specified data ID
        """
        if hasattr(self, f"get_{data_id}"):
            return getattr(self, f"get_{data_id}")(**kwargs)
        raise ValueError(
            f"{self.__class__.__name__} doesn't have getter for '{data_id}' auxiliary data. Available data: {self.get_aux_data_ids()}"
        )


class CsvStorageDataReader(DataReader):
    """
    Data reader for timeseries data stored as csv files in the specified directory
    """

    def __init__(self, path: str) -> None:
        if not exists(path):
            raise ValueError(f"Folder is not found at {path}")
        self.path = path

    def __find_time_idx(self, arr: pa.ChunkedArray, v) -> int:
        ix = arr.index(v).as_py()
        if ix < 0:
            for c in arr.iterchunks():
                a = c.to_numpy()
                ix = np.searchsorted(a, v, side="right")
                if ix > 0 and ix < len(c):
                    ix = arr.index(a[ix]).as_py() - 1
                    break
        return ix

    def __check_file_name(self, name: str) -> str | None:
        _f = join(self.path, name.replace(":", os.sep))
        for sfx in [".csv", ".csv.gz", ""]:
            if exists(p := (_f + sfx)):
                return p
        return None

    def read(
        self,
        data_id: str,
        start: str | None = None,
        stop: str | None = None,
        transform: DataTransformer = DataTransformer(),
        chunksize=0,
        timestamp_formatters=None,
        timeframe=None,
        **kwargs,
    ) -> Iterable | Any:

        f_path = self.__check_file_name(data_id)
        if not f_path:
            ValueError(f"Can't find any csv data for {data_id} in {self.path} !")

        convert_options = None
        if timestamp_formatters is not None:
            convert_options = csv.ConvertOptions(timestamp_parsers=timestamp_formatters)

        table = csv.read_csv(
            f_path,
            parse_options=csv.ParseOptions(ignore_empty_lines=True),
            convert_options=convert_options,
        )
        fieldnames = table.column_names

        # - try to find range to load
        start_idx, stop_idx = 0, table.num_rows
        try:
            _time_field_idx = _FIND_TIME_COL_IDX(fieldnames)
            _time_type = table.field(_time_field_idx).type
            _time_unit = _time_type.unit if hasattr(_time_type, "unit") else "ms"
            _time_data = table[_time_field_idx]

            # - check if need convert time to primitive types (i.e. Date32 -> timestamp[x])
            _time_cast_function = lambda xs: xs
            if _time_type != pa.timestamp(_time_unit):
                _time_cast_function = lambda xs: xs.cast(pa.timestamp(_time_unit))
                _time_data = _time_cast_function(_time_data)

            # - preprocessing start and stop
            t_0, t_1 = handle_start_stop(start, stop, convert=lambda x: _recognize_t(x, None, _time_unit))

            # - check requested range
            if t_0:
                start_idx = self.__find_time_idx(_time_data, t_0)
                if start_idx >= table.num_rows:
                    # no data for requested start date
                    return None

            if t_1:
                stop_idx = self.__find_time_idx(_time_data, t_1)
                if stop_idx < 0 or stop_idx < start_idx:
                    stop_idx = table.num_rows

        except Exception as exc:
            logger.warning(exc)
            logger.info("loading whole file")

        length = stop_idx - start_idx + 1
        selected_table = table.slice(start_idx, length)

        # - in this case we want to return iterable chunks of data
        if chunksize > 0:

            def _iter_chunks():
                for n in range(0, length // chunksize + 1):
                    transform.start_transform(data_id, fieldnames, start=start, stop=stop)
                    raw_data = selected_table[n * chunksize : min((n + 1) * chunksize, length)].to_pandas().to_numpy()
                    transform.process_data(raw_data)
                    yield transform.collect()

            return _iter_chunks()

        transform.start_transform(data_id, fieldnames, start=start, stop=stop)
        raw_data = selected_table.to_pandas().to_numpy()
        transform.process_data(raw_data)
        return transform.collect()

    def get_candles(
        self,
        exchange: str,
        symbols: list[str],
        start: str | pd.Timestamp,
        stop: str | pd.Timestamp,
        timeframe: str | None = None,
    ) -> pd.DataFrame:
        """
        Returns pandas DataFrame of candles for given exchange and symbols within specified time range and timeframe
        """
        _r = []
        for symbol in symbols:
            x = self.read(
                f"{exchange}:{symbol}", start=start, stop=stop, timeframe=timeframe, transform=AsPandasFrame()
            )
            if x is not None:
                if timeframe is not None:
                    x = ohlc_resample(x, timeframe)
                _r.append(x.assign(symbol=symbol.upper(), timestamp=x.index))  # type: ignore
        return srows(*_r).set_index(["timestamp", "symbol"])

    def get_names(self, **kwargs) -> List[str]:
        _n = []
        for root, _, files in os.walk(self.path):
            path = root.split(os.sep)
            for file in files:
                if m := re.match(r"(.*)\.csv(.gz)?$", file):
                    f = path[-1]
                    n = file.split(".")[0]
                    if f == self.path:
                        name = n
                    else:
                        name = f"{f}:{ n }" if f else n
                    _n.append(name)
        return _n


class InMemoryDataFrameReader(DataReader):
    """
    Data reader for pandas DataFrames
    """

    exchange: str | None
    _data: Dict[str, pd.DataFrame | pd.Series]

    def __init__(self, data: Dict[str, pd.DataFrame | pd.Series], exchange: str | None = None) -> None:
        if not isinstance(data, dict):
            raise ValueError("data must be a dictionary of pandas DataFrames")
        self._data = data
        self.exchange = exchange

    def get_names(self, **kwargs) -> List[str]:
        keys = list(self._data.keys())
        if self.exchange:
            return [f"{self.exchange}:{k}" for k in keys]
        return keys

    def read(
        self,
        data_id: str,
        start: str | None = None,
        stop: str | None = None,
        transform: DataTransformer = DataTransformer(),
        chunksize=0,
        **kwargs,
    ) -> Iterable | List:
        """
        Read and transform data for a given data_id within a specified time range.

        Parameters:
        -----------
        data_id : str
            The identifier for the data to be read.
        start : str | None, optional
            The start time for the data range (inclusive). If None, start from the earliest available data.
        stop : str | None, optional
            The stop time for the data range (inclusive). If None, include data up to the latest available.
        transform : DataTransformer, optional
            An instance of DataTransformer to process the retrieved data. Defaults to DataTransformer().
        chunksize : int, optional
            The size of data chunks to process at a time. If 0, process all data at once. Defaults to 0.
        **kwargs : dict
            Additional keyword arguments for future extensions.

        Returns:
        --------
        Iterable | List
            The processed and transformed data, either as an iterable (if chunksize > 0) or as a list.

        Raises:
        -------
        ValueError
            If no data is found for the given data_id.
        """
        start, stop = handle_start_stop(start, stop)
        if data_id not in self._data:
            if data_id.startswith(self.exchange):
                data_id = data_id.split(":")[1]
        d = self._data.get(data_id)
        if d is None:
            raise ValueError(f"No data found for {data_id}")
        d2 = d.loc[start:stop].copy().reset_index()
        transform.start_transform(data_id, list(d2.columns), start=start, stop=stop)
        transform.process_data(d2.values)
        res = transform.collect()
        if chunksize > 0:

            def __iterable():
                yield res

            return __iterable()
        return res

    def get_symbols(self, **kwargs) -> List[str]:
        return self.get_names()


class AsPandasFrame(DataTransformer):
    """
    List of records to pandas dataframe transformer
    """

    def __init__(self, timestamp_units=None) -> None:
        self.timestamp_units = timestamp_units

    def start_transform(self, name: str, column_names: List[str], **kwargs):
        self._time_idx = _FIND_TIME_COL_IDX(column_names)
        self._column_names = column_names
        self._frame = pd.DataFrame()

    def process_data(self, rows_data: Iterable) -> Any:
        self._frame
        p = pd.DataFrame.from_records(rows_data, columns=self._column_names)
        p.set_index(self._column_names[self._time_idx], drop=True, inplace=True)
        p.index = pd.to_datetime(p.index, unit=self.timestamp_units) if self.timestamp_units else p.index
        p.index.rename("timestamp", inplace=True)
        p.sort_index(inplace=True)
        self._frame = pd.concat((self._frame, p), axis=0, sort=True)
        return p

    def collect(self) -> Any:
        return self._frame


class AsOhlcvSeries(DataTransformer):
    """
    Convert incoming data into OHLCV series.

    Incoming data may have one of the following structures:

        ```
        ohlcv:        time,open,high,low,close,volume|quote_volume,(buy_volume)
        quotes:       time,bid,ask,bidsize,asksize
        trades (TAS): time,price,size,(is_taker)
        ```
    """

    def __init__(self, timeframe: str | None = None, timestamp_units="ns") -> None:
        super().__init__()
        self.timeframe = timeframe
        self._series = None
        self._data_type = None
        self.timestamp_units = timestamp_units

    def start_transform(self, name: str, column_names: List[str], **kwargs):
        self._time_idx = _FIND_TIME_COL_IDX(column_names)
        self._volume_idx = None
        self._b_volume_idx = None
        try:
            self._close_idx = _find_column_index_in_list(column_names, "close")
            self._open_idx = _find_column_index_in_list(column_names, "open")
            self._high_idx = _find_column_index_in_list(column_names, "high")
            self._low_idx = _find_column_index_in_list(column_names, "low")

            try:
                self._volume_idx = _find_column_index_in_list(column_names, "quote_volume", "volume", "vol")
            except:
                pass

            try:
                self._b_volume_idx = _find_column_index_in_list(
                    column_names,
                    "taker_buy_volume",
                    "taker_buy_quote_volume",
                    "buy_volume",
                )
            except:
                pass

            self._data_type = "ohlc"
        except:
            try:
                self._ask_idx = _find_column_index_in_list(column_names, "ask")
                self._bid_idx = _find_column_index_in_list(column_names, "bid")
                self._data_type = "quotes"
            except:

                try:
                    self._price_idx = _find_column_index_in_list(column_names, "price")
                    self._size_idx = _find_column_index_in_list(
                        column_names, "quote_qty", "qty", "size", "amount", "volume"
                    )
                    self._taker_idx = None
                    try:
                        self._taker_idx = _find_column_index_in_list(
                            column_names,
                            "is_buyer_maker",
                            "side",
                            "aggressive",
                            "taker",
                            "is_taker",
                        )
                    except:
                        pass

                    self._data_type = "trades"
                except:
                    raise ValueError(f"Can't recognize data for update from header: {column_names}")

        self._column_names = column_names
        self._name = name
        if self.timeframe:
            self._series = OHLCV(self._name, self.timeframe)

    def _proc_ohlc(self, rows_data: List[List]):
        for d in rows_data:
            self._series.update_by_bar(
                _time(d[self._time_idx], self.timestamp_units),
                d[self._open_idx],
                d[self._high_idx],
                d[self._low_idx],
                d[self._close_idx],
                d[self._volume_idx] if self._volume_idx else 0,
                d[self._b_volume_idx] if self._b_volume_idx else 0,
            )

    def _proc_quotes(self, rows_data: List[List]):
        for d in rows_data:
            self._series.update(
                _time(d[self._time_idx], self.timestamp_units),
                (d[self._ask_idx] + d[self._bid_idx]) / 2,
            )

    def _proc_trades(self, rows_data: List[List]):
        for d in rows_data:
            a = d[self._taker_idx] if self._taker_idx else 0
            s = d[self._size_idx]
            b = s if a else 0
            self._series.update(_time(d[self._time_idx], self.timestamp_units), d[self._price_idx], s, b)

    def process_data(self, rows_data: List[List]) -> Any:
        if self._series is None:
            ts = [t[self._time_idx] for t in rows_data[:100]]
            self.timeframe = pd.Timedelta(infer_series_frequency(ts)).asm8.item()

            # - create instance after first data received if
            self._series = OHLCV(self._name, self.timeframe)

        match self._data_type:
            case "ohlc":
                self._proc_ohlc(rows_data)
            case "quotes":
                self._proc_quotes(rows_data)
            case "trades":
                self._proc_trades(rows_data)

        return None

    def collect(self) -> Any:
        return self._series


class AsQuotes(DataTransformer):
    """
    Tries to convert incoming data to list of Quote's
    Data must have appropriate structure: bid, ask, bidsize, asksize and time
    """

    def start_transform(self, name: str, column_names: List[str], **kwargs):
        self.buffer = list()
        self._time_idx = _FIND_TIME_COL_IDX(column_names)
        self._bid_idx = _find_column_index_in_list(column_names, "bid")
        self._ask_idx = _find_column_index_in_list(column_names, "ask")
        self._bidvol_idx = _find_column_index_in_list(column_names, "bidvol", "bid_vol", "bidsize", "bid_size")
        self._askvol_idx = _find_column_index_in_list(column_names, "askvol", "ask_vol", "asksize", "ask_size")

    def process_data(self, rows_data: Iterable) -> Any:
        if rows_data is not None:
            for d in rows_data:
                t = d[self._time_idx]
                b = d[self._bid_idx]
                a = d[self._ask_idx]
                bv = d[self._bidvol_idx]
                av = d[self._askvol_idx]
                self.buffer.append(Quote(_time(t, "ns"), b, a, bv, av))


class AsTrades(DataTransformer):
    """
    Tries to convert incoming data to list of Trades
    Data must have appropriate structure: price, size, market_maker (optional).
    Market maker column specifies if buyer is a maker or taker.
    """

    def start_transform(self, name: str, column_names: List[str], **kwargs):
        self.buffer: list[Trade] = list()
        self._time_idx = _FIND_TIME_COL_IDX(column_names)
        self._price_idx = _find_column_index_in_list(column_names, "price")
        self._size_idx = _find_column_index_in_list(column_names, "size")
        try:
            self._side_idx = _find_column_index_in_list(column_names, "market_maker")
        except:
            self._side_idx = None

    def process_data(self, rows_data: Iterable) -> Any:
        if rows_data is not None:
            for d in rows_data:
                t = d[self._time_idx]
                price = d[self._price_idx]
                size = d[self._size_idx]
                side = d[self._side_idx] if self._side_idx else -1
                self.buffer.append(Trade(_time(t, "ns"), price, size, side))


class AsTimestampedRecords(DataTransformer):
    """
    Convert incoming data to list or dictionaries with preprocessed timestamps ('timestamp_ns' and 'timestamp')
    ```
    [
        {
            'open_time': 1711944240000.0,
            'open': 203.219,
            'high': 203.33,
            'low': 203.134,
            'close': 203.175,
            'volume': 10060.0,
            ....
            'timestamp_ns': 1711944240000000000,
            'timestamp': Timestamp('2024-04-01 04:04:00')
        },
        ...
    ] ```
    """

    def __init__(self, timestamp_units: str | None = None) -> None:
        self.timestamp_units = timestamp_units

    def start_transform(self, name: str, column_names: List[str], **kwargs):
        self.buffer = list()
        self._time_idx = _FIND_TIME_COL_IDX(column_names)
        self._column_names = column_names

    def process_data(self, rows_data: Iterable) -> Any:
        self.buffer.extend(rows_data)

    def collect(self) -> Any:
        res = []
        for r in self.buffer:
            t = r[self._time_idx]
            if self.timestamp_units:
                t = _time(t, self.timestamp_units)
            di = dict(zip(self._column_names, r)) | {
                "timestamp_ns": t,
                "timestamp": pd.Timestamp(t),
            }
            res.append(di)
        return res


class RestoreTicksFromOHLC(DataTransformer):
    """
    Emulates quotes (and trades) from OHLC bars
    """

    def __init__(
        self,
        trades: bool = False,  # if we also wants 'trades'
        default_bid_size=1e9,  # default bid/ask is big
        default_ask_size=1e9,  # default bid/ask is big
        daily_session_start_end=DEFAULT_DAILY_SESSION,
        timestamp_units="ns",
        spread=0.0,
    ):
        super().__init__()
        self._trades = trades
        self._bid_size = default_bid_size
        self._ask_size = default_ask_size
        self._s2 = spread / 2.0
        self._d_session_start = daily_session_start_end[0]
        self._d_session_end = daily_session_start_end[1]
        self._timestamp_units = timestamp_units

    def start_transform(self, name: str, column_names: List[str], **kwargs):
        self.buffer = []
        # - it will fail if receive data doesn't look as ohlcv
        self._time_idx = _FIND_TIME_COL_IDX(column_names)
        self._open_idx = _find_column_index_in_list(column_names, "open")
        self._high_idx = _find_column_index_in_list(column_names, "high")
        self._low_idx = _find_column_index_in_list(column_names, "low")
        self._close_idx = _find_column_index_in_list(column_names, "close")
        self._volume_idx = None
        self._freq = None
        try:
            self._volume_idx = _find_column_index_in_list(column_names, "volume", "vol")
        except:
            pass

        if self._volume_idx is None and self._trades:
            logger.warning("Input OHLC data doesn't contain volume information so trades can't be emulated !")
            self._trades = False

    def process_data(self, rows_data: List[List]) -> Any:
        if rows_data is None:
            return

        s2 = self._s2

        if self._freq is None:
            ts = [t[self._time_idx] for t in rows_data[:100]]
            try:
                self._freq = infer_series_frequency(ts)
            except ValueError:
                logger.warning("Can't determine frequency of incoming data")
                return

            # - timestamps when we emit simulated quotes
            dt = self._freq.astype("timedelta64[ns]").item()
            if dt < D1:
                self._t_start = MS1  # dt // 10
                self._t_mid1 = dt // 2 - dt // 10
                self._t_mid2 = dt // 2 + dt // 10
                self._t_end = dt - MS1  # dt - dt // 10
            else:
                self._t_start = self._d_session_start
                self._t_mid1 = dt // 2 - H1
                self._t_mid2 = dt // 2 + H1
                self._t_end = self._d_session_end

        # - input data
        for data in rows_data:
            # ti = pd.Timestamp(data[self._time_idx]).as_unit("ns").asm8.item()
            ti = _time(data[self._time_idx], self._timestamp_units)
            o = data[self._open_idx]
            h = data[self._high_idx]
            l = data[self._low_idx]
            c = data[self._close_idx]
            rv = data[self._volume_idx] if self._volume_idx else 0

            # - opening quote
            self.buffer.append(Quote(ti + self._t_start, o - s2, o + s2, self._bid_size, self._ask_size))

            if c >= o:
                if self._trades:
                    self.buffer.append(Trade(ti + self._t_start, o - s2, rv * (o - l)))  # sell 1
                self.buffer.append(
                    Quote(
                        ti + self._t_mid1,
                        l - s2,
                        l + s2,
                        self._bid_size,
                        self._ask_size,
                    )
                )

                if self._trades:
                    self.buffer.append(Trade(ti + self._t_mid1, l + s2, rv * (c - o)))  # buy 1
                self.buffer.append(
                    Quote(
                        ti + self._t_mid2,
                        h - s2,
                        h + s2,
                        self._bid_size,
                        self._ask_size,
                    )
                )

                if self._trades:
                    self.buffer.append(Trade(ti + self._t_mid2, h - s2, rv * (h - c)))  # sell 2
            else:
                if self._trades:
                    self.buffer.append(Trade(ti + self._t_start, o + s2, rv * (h - o)))  # buy 1
                self.buffer.append(
                    Quote(
                        ti + self._t_mid1,
                        h - s2,
                        h + s2,
                        self._bid_size,
                        self._ask_size,
                    )
                )

                if self._trades:
                    self.buffer.append(Trade(ti + self._t_mid1, h - s2, rv * (o - c)))  # sell 1
                self.buffer.append(
                    Quote(
                        ti + self._t_mid2,
                        l - s2,
                        l + s2,
                        self._bid_size,
                        self._ask_size,
                    )
                )

                if self._trades:
                    self.buffer.append(Trade(ti + self._t_mid2, l + s2, rv * (c - l)))  # buy 2

            # - closing quote
            self.buffer.append(Quote(ti + self._t_end, c - s2, c + s2, self._bid_size, self._ask_size))


def _retry(fn):
    @wraps(fn)
    def wrapper(*args, **kw):
        cls = args[0]
        for x in range(cls._reconnect_tries):
            # print(x, cls._reconnect_tries)
            try:
                return fn(*args, **kw)
            except (pg.InterfaceError, pg.OperationalError, AttributeError) as e:
                logger.debug("Database Connection [InterfaceError or OperationalError]")
                # print ("Idle for %s seconds" % (cls._reconnect_idle))
                # time.sleep(cls._reconnect_idle)
                cls._connect()

    return wrapper


class QuestDBSqlBuilder:
    """
    Generic sql builder for QuestDB data
    """

    def get_table_name(self, data_id: str, sfx: str = "") -> str | None:
        pass

    def prepare_data_sql(
        self,
        data_id: str,
        start: str | None,
        end: str | None,
        resample: str | None,
        data_type: str,
    ) -> str | None:
        pass

    def prepare_names_sql(self) -> str:
        return "select table_name from tables()"


class QuestDBSqlCandlesBuilder(QuestDBSqlBuilder):
    """
    Sql builder for candles data
    """

    def get_table_name(self, data_id: str, sfx: str = "") -> str:
        """
        Get table name for data_id
        data_id can have format <exchange>.<type>:<symbol>
        for example:
            BINANCE.UM:BTCUSDT or BINANCE:BTCUSDT for spot
        """
        _aliases = {"um": "umfutures", "cm": "cmfutures", "f": "futures"}
        sfx = sfx or "candles_1m"
        table_name = data_id
        _ss = data_id.split(":")
        if len(_ss) > 1:
            _exch, symb = _ss
            _mktype = "spot"
            _ss = _exch.split(".")
            if len(_ss) > 1:
                _exch = _ss[0]
                _mktype = _ss[1]
            _mktype = _mktype.lower()
            parts = [_exch.lower(), _aliases.get(_mktype, _mktype)]
            if "candles" not in sfx:
                parts.append(symb.lower())
            parts.append(sfx)
            table_name = ".".join(filter(lambda x: x, parts))
        return table_name

    def prepare_names_sql(self) -> str:
        return "select table_name from tables() where table_name like '%candles%'"

    @staticmethod
    def _convert_time_delta_to_qdb_resample_format(c_tf: str):
        if c_tf:
            _t = re.match(r"(\d+)(\w+)", c_tf)
            if _t and len(_t.groups()) > 1:
                c_tf = f"{_t[1]}{_t[2][0].lower()}"
        return c_tf

    def prepare_data_sql(
        self,
        data_id: str,
        start: str | None,
        end: str | None,
        resample: str | None,
        data_type: str,
    ) -> str:
        _ss = data_id.split(":")
        if len(_ss) > 1:
            _exch, symb = _ss
        else:
            symb = data_id

        symb = symb.lower()

        where = f"where symbol = '{symb}'"
        w0 = f"timestamp >= '{start}'" if start else ""
        w1 = f"timestamp < '{end}'" if end else ""

        # - fix: when no data ranges are provided we must skip empy where keyword
        if w0 or w1:
            where = f"{where} and {w0} and {w1}" if (w0 and w1) else f"{where} and {(w0 or w1)}"

        # - filter out candles without any volume
        where = f"{where} and volume > 0"

        # - check resample format
        resample = (
            QuestDBSqlCandlesBuilder._convert_time_delta_to_qdb_resample_format(resample)
            if resample
            else "1m"  # if resample is empty let's use 1 minute timeframe
        )
        _rsmpl = f"SAMPLE by {resample} FILL(NONE)" if resample else ""

        table_name = self.get_table_name(data_id, data_type)
        return f"""
                select timestamp, 
                first(open) as open, 
                max(high) as high,
                min(low) as low,
                last(close) as close,
                sum(volume) as volume,
                sum(quote_volume) as quote_volume,
                sum(count) as count,
                sum(taker_buy_volume) as taker_buy_volume,
                sum(taker_buy_quote_volume) as taker_buy_quote_volume
                from "{table_name}" {where} {_rsmpl};
            """


class QuestDBConnector(DataReader):
    """
    Very first version of QuestDB connector

    ### Connect to an existing QuestDB instance
    >>> db = QuestDBConnector()
    >>> db.read('BINANCE.UM:ETHUSDT', '2024-01-01', transform=AsPandasFrame())
    """

    _reconnect_tries = 5
    _reconnect_idle = 0.1  # wait seconds before retying
    _builder: QuestDBSqlBuilder

    def __init__(
        self,
        builder: QuestDBSqlBuilder = QuestDBSqlCandlesBuilder(),
        host="localhost",
        user="admin",
        password="quest",
        port=8812,
    ) -> None:
        self._connection = None
        self._host = host
        self._port = port
        self.connection_url = f"user={user} password={password} host={host} port={port}"
        self._builder = builder
        self._connect()

    def __getstate__(self):
        if self._connection:
            self._connection.close()
            self._connection = None
        state = self.__dict__.copy()
        return state

    def _connect(self):
        self._connection = pg.connect(self.connection_url, autocommit=True)
        logger.debug(f"Connected to QuestDB at {self._host}:{self._port}")

    def read(
        self,
        data_id: str,
        start: str | None = None,
        stop: str | None = None,
        transform: DataTransformer = DataTransformer(),
        chunksize=0,
        timeframe: str | None = "1m",
        data_type="candles_1m",
    ) -> Any:
        return self._read(
            data_id,
            start,
            stop,
            transform,
            chunksize,
            timeframe,
            data_type,
            self._builder,
        )

    def get_symbols(self, exchange: str) -> list[str]:
        table_name = QuestDBSqlCandlesBuilder().get_table_name(f"{exchange}:BTCUSDT")
        query = f"""
        select distinct symbol
        from "{table_name}";
        """
        return self.execute(query)["symbol"].tolist()

    def get_candles(
        self,
        exchange: str,
        symbols: list[str],
        start: str | pd.Timestamp,
        stop: str | pd.Timestamp,
        timeframe: str = "1d",
    ) -> pd.DataFrame:
        assert len(symbols) > 0, "No symbols provided"
        quoted_symbols = [f"'{s.lower()}'" for s in symbols]
        where = f"where symbol in ({', '.join(quoted_symbols)}) and timestamp >= '{start}' and timestamp < '{stop}'"
        table_name = QuestDBSqlCandlesBuilder().get_table_name(f"{exchange}:{symbols[0]}")

        _rsmpl = f"sample by {QuestDBSqlCandlesBuilder._convert_time_delta_to_qdb_resample_format(timeframe)}"

        query = f"""
        select timestamp, 
        upper(symbol) as symbol,
        first(open) as open, 
        max(high) as high,
        min(low) as low,
        last(close) as close,
        sum(volume) as volume,
        sum(quote_volume) as quote_volume,
        sum(count) as count,
        sum(taker_buy_volume) as taker_buy_volume,
        sum(taker_buy_quote_volume) as taker_buy_quote_volume
        from "{table_name}" {where} {_rsmpl};
        """
        res = self.execute(query)
        if res.empty:
            return res
        return res.set_index(["timestamp", "symbol"])

    def get_average_quote_volume(
        self,
        exchange: str,
        start: str | pd.Timestamp,
        stop: str | pd.Timestamp,
        timeframe: str = "1d",
    ) -> pd.Series:
        table_name = QuestDBSqlCandlesBuilder().get_table_name(f"{exchange}:BTCUSDT")
        query = f"""
        WITH sampled as (
            select timestamp, symbol, sum(quote_volume) as qvolume 
            from "{table_name}"
            where timestamp >= '{start}' and timestamp < '{stop}'
            SAMPLE BY {QuestDBSqlCandlesBuilder._convert_time_delta_to_qdb_resample_format(timeframe)}
        )
        select upper(symbol) as symbol, avg(qvolume) as quote_volume from sampled
        group by symbol
        order by quote_volume desc;
        """
        vol_stats = self.execute(query)
        if vol_stats.empty:
            return pd.Series()
        return vol_stats.set_index("symbol")["quote_volume"]

    def get_fundamental_data(
        self, exchange: str, start: str | pd.Timestamp | None = None, stop: str | pd.Timestamp | None = None
    ) -> pd.DataFrame:
        table_name = {"BINANCE.UM": "binance.umfutures.fundamental"}[exchange]
        query = f"select * from {table_name}"
        if start or stop:
            conditions = []
            if start:
                conditions.append(f"timestamp >= '{start}'")
            if stop:
                conditions.append(f"timestamp < '{stop}'")
            query += " where " + " and ".join(conditions)
        df = self.execute(query)
        if df.empty:
            return pd.DataFrame()
        return df.set_index(["timestamp", "symbol", "metric"]).value.unstack("metric")

    def get_names(self) -> List[str]:
        return self._get_names(self._builder)

    @_retry
    def execute(self, query: str) -> pd.DataFrame:
        _cursor = self._connection.cursor()  # type: ignore
        _cursor.execute(query)  # type: ignore
        names = [d.name for d in _cursor.description]  # type: ignore
        records = _cursor.fetchall()
        if not records:
            return pd.DataFrame()
        return pd.DataFrame(records, columns=names)

    @_retry
    def _read(
        self,
        data_id: str,
        start: str | None,
        stop: str | None,
        transform: DataTransformer,
        chunksize: int,
        timeframe: str | None,
        data_type: str,
        builder: QuestDBSqlBuilder,
    ) -> Any:
        start, end = handle_start_stop(start, stop)
        _req = builder.prepare_data_sql(data_id, start, end, timeframe, data_type)

        _cursor = self._connection.cursor()  # type: ignore
        _cursor.execute(_req)  # type: ignore
        names = [d.name for d in _cursor.description]  # type: ignore

        if chunksize > 0:

            def _iter_chunks():
                while True:
                    records = _cursor.fetchmany(chunksize)
                    if not records:
                        _cursor.close()
                        break
                    transform.start_transform(data_id, names, start=start, stop=stop)
                    transform.process_data(records)
                    yield transform.collect()

            return _iter_chunks()

        try:
            records = _cursor.fetchall()
            if not records:
                return None
            transform.start_transform(data_id, names, start=start, stop=stop)
            transform.process_data(records)
            return transform.collect()
        finally:
            _cursor.close()

    @_retry
    def _get_names(self, builder: QuestDBSqlBuilder) -> List[str]:
        _cursor = None
        try:
            _cursor = self._connection.cursor()  # type: ignore
            _cursor.execute(builder.prepare_names_sql())  # type: ignore
            records = _cursor.fetchall()
        finally:
            if _cursor:
                _cursor.close()
        return [r[0] for r in records]

    def __del__(self):
        try:
            if self._connection is not None:
                logger.debug("Closing connection")
                self._connection.close()
        except:
            pass


class QuestDBSqlOrderBookBuilder(QuestDBSqlCandlesBuilder):
    """
    Sql builder for snapshot data
    """

    SNAPSHOT_DELTA = pd.Timedelta("1h")
    MIN_DELTA = pd.Timedelta("1s")

    def prepare_data_sql(
        self,
        data_id: str,
        start: str | None,
        end: str | None,
        resample: str,
        data_type: str,
    ) -> str:
        if not start or not end:
            raise ValueError("Start and end dates must be provided for orderbook data!")
        start_dt, end_dt = pd.Timestamp(start), pd.Timestamp(end)
        delta = end_dt - start_dt

        raw_start_dt = start_dt.floor(self.SNAPSHOT_DELTA) - self.MIN_DELTA

        table_name = self.get_table_name(data_id, data_type)
        query = f"""
SELECT * FROM {table_name}
WHERE timestamp BETWEEN '{raw_start_dt}' AND '{end_dt}'
"""
        return query


class TradeSql(QuestDBSqlCandlesBuilder):

    def prepare_data_sql(
        self,
        data_id: str,
        start: str | None,
        end: str | None,
        resample: str,
        data_type: str,
    ) -> str:
        table_name = self.get_table_name(data_id, data_type)
        where = ""
        w0 = f"timestamp >= '{start}'" if start else ""
        w1 = f"timestamp <= '{end}'" if end else ""

        # - fix: when no data ranges are provided we must skip empy where keyword
        if w0 or w1:
            where = f"where {w0} and {w1}" if (w0 and w1) else f"where {(w0 or w1)}"

        resample = (
            QuestDBSqlCandlesBuilder._convert_time_delta_to_qdb_resample_format(resample) if resample else resample
        )
        if resample:
            sql = f"""
                select timestamp, first(price) as open, max(price) as high, min(price) as low, last(price) as close, 
                sum(size) as volume from "{table_name}" {where} SAMPLE by {resample};"""
        else:
            sql = f"""select timestamp, price, size, market_maker from "{table_name}" {where};"""

        return sql


class MultiQdbConnector(QuestDBConnector):
    """
    Data connector for QuestDB which provides access to following data types:
      - candles
      - trades
      - orderbook snapshots
      - liquidations
      - funding rate

    Examples:
    1. Retrieving trades:
        qdb.read(
            "BINANCE.UM:BTCUSDT",
            "2023-01-01 00:00",
            "2023-01-01 10:00",
            timeframe="15Min",
            transform=AsPandasFrame(),
            data_type="trade"
        )
    """

    _TYPE_TO_BUILDER = {
        "candles_1m": QuestDBSqlCandlesBuilder(),
        "trade": TradeSql(),
        "agg_trade": TradeSql(),
        "orderbook": QuestDBSqlOrderBookBuilder(),
    }

    _TYPE_MAPPINGS = {
        "candles": "candles_1m",
        "ohlc": "candles_1m",
        "trades": "trade",
        "ob": "orderbook",
        "trd": "trade",
        "td": "trade",
        "aggTrade": "agg_trade",
        "agg_trades": "agg_trade",
        "aggTrades": "agg_trade",
    }

    def __init__(
        self,
        host="localhost",
        user="admin",
        password="quest",
        port=8812,
    ) -> None:
        self._connection = None
        self._host = host
        self._port = port
        self._user = user
        self._password = password
        self._connect()

    @property
    def connection_url(self):
        return " ".join(
            [
                f"user={self._user}",
                f"password={self._password}",
                f"host={self._host}",
                f"port={self._port}",
            ]
        )

    def read(
        self,
        data_id: str,
        start: str | None = None,
        stop: str | None = None,
        transform: DataTransformer = DataTransformer(),
        chunksize: int = 0,
        timeframe: str | None = None,
        data_type: str = "candles",
    ) -> Any:
        _mapped_data_type = self._TYPE_MAPPINGS.get(data_type, data_type)
        return self._read(
            data_id,
            start,
            stop,
            transform,
            chunksize,
            timeframe,
            _mapped_data_type,
            self._TYPE_TO_BUILDER[_mapped_data_type],
        )

    def get_names(self, data_type: str) -> List[str]:
        return self._get_names(self._TYPE_TO_BUILDER[self._TYPE_MAPPINGS.get(data_type, data_type)])



import pandas as pd
import heapq

from dataclasses import dataclass
from collections import defaultdict
from typing import Any, Iterator, Iterable
from concurrent.futures import ProcessPoolExecutor, ThreadPoolExecutor, Future

from qubx import logger
from qubx.core.basics import Instrument, dt_64, BatchEvent
from qubx.data.readers import DataReader, DataTransformer
from qubx.utils.misc import Stopwatch
from qubx.core.exceptions import SimulatorError


_SW = Stopwatch()


class DataLoader:
    _TYPE_MAPPERS = {"agg_trade": "trade", "ohlc": "bar", "ohlcv": "bar"}

    def __init__(
        self,
        transformer: DataTransformer,
        reader: DataReader,
        instrument: Instrument,
        timeframe: str | None,
        preload_bars: int = 0,
        data_type: str = "ohlc",
        output_type: str | None = None,  # transfomer can somtimes map to a different output type
        chunksize: int = 5_000,
    ) -> None:
        self._instrument = instrument
        self._spec = f"{instrument.exchange}:{instrument.symbol}"
        self._reader = reader
        self._transformer = transformer
        self._init_bars_required = preload_bars
        self._timeframe = timeframe
        self._data_type = data_type
        self._output_type = output_type
        self._first_load = True
        self._chunksize = chunksize

    def load(self, start: str | pd.Timestamp, end: str | pd.Timestamp) -> Iterator:
        if self._first_load:
            if self._init_bars_required > 0 and self._timeframe:
                start = pd.Timestamp(start) - self._init_bars_required * pd.Timedelta(self._timeframe)
            self._first_load = False

        args = dict(
            data_id=self._spec,
            start=start,
            stop=end,
            transform=self._transformer,
            data_type=self._data_type,
            chunksize=self._chunksize,
        )

        if self._timeframe:
            args["timeframe"] = self._timeframe

        return self._reader.read(**args)  # type: ignore

    @property
    def instrument(self) -> Instrument:
        return self._instrument

    @property
    def symbol(self) -> str:
        return self._instrument.symbol

    @property
    def data_type(self) -> str:
        if self._output_type:
            return self._output_type
        return self._TYPE_MAPPERS.get(self._data_type, self._data_type)

    def __hash__(self) -> int:
        return hash((self._instrument.symbol, self._data_type))

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, DataLoader):
            return False
        return self._instrument.symbol == other._instrument.symbol and self._data_type == other._data_type


class SimulatedDataQueue:
    _loaders: dict[str, list[DataLoader]]

    def __init__(self):
        self._loaders = defaultdict(list)
        self._start = None
        self._stop = None
        self._current_time = None
        self._index_to_loader: dict[int, DataLoader] = {}
        self._loader_to_index = {}
        self._latest_loader_index = -1
        self._removed_loader_indices = set()

    @property
    def is_running(self) -> bool:
        return self._current_time is not None

    def __add__(self, loader: DataLoader) -> "SimulatedDataQueue":
        self._latest_loader_index += 1
        new_loader_index = self._latest_loader_index
        self._loaders[loader.symbol].append(loader)
        self._index_to_loader[new_loader_index] = loader
        self._loader_to_index[loader] = new_loader_index
        if self.is_running:
            self._add_chunk_to_heap(new_loader_index)
        return self

    def __sub__(self, loader: DataLoader) -> "SimulatedDataQueue":
        loader_index = self._loader_to_index[loader]
        self._loaders[loader.symbol].remove(loader)
        del self._index_to_loader[loader_index]
        del self._loader_to_index[loader]
        del self._index_to_chunk_size[loader_index]
        del self._index_to_iterator[loader_index]
        self._removed_loader_indices.add(loader_index)
        return self

    def get_loader(self, symbol: str, data_type: str) -> DataLoader:
        loaders = self._loaders[symbol]
        for loader in loaders:
            if loader.data_type == data_type:
                return loader
        raise ValueError(f"Loader for {symbol} and {data_type} not found")

    def create_iterable(self, start: str | pd.Timestamp, stop: str | pd.Timestamp) -> Iterator:
        self._start = start
        self._stop = stop
        self._current_time = None
        return self

    def __iter__(self) -> Iterator:
        logger.debug("Initializing chunks for each loader")
        assert self._start is not None
        self._current_time = int(pd.Timestamp(self._start).timestamp() * 1e9)
        self._index_to_chunk_size = {}
        self._index_to_iterator = {}
        self._event_heap = []
        for loader_index in self._index_to_loader.keys():
            self._add_chunk_to_heap(loader_index)
        return self

    @_SW.watch("DataQueue")
    def __next__(self) -> tuple[str, str, Any]:
        if not self._event_heap:
            raise StopIteration

        loader_index = None

        # get the next event from the heap
        # if the loader_index is in the removed_loader_indices, skip it (optimization to avoid unnecessary heap operations)
        while self._event_heap and (loader_index is None or loader_index in self._removed_loader_indices):
            dt, loader_index, chunk_index, event = heapq.heappop(self._event_heap)

        if loader_index is None or loader_index in self._removed_loader_indices:
            raise StopIteration

        loader = self._index_to_loader[loader_index]
        data_type = loader.data_type
        if dt < self._current_time:  # type: ignore
            data_type = f"hist_{data_type}"
        else:
            # only update the current time if the event is not historical
            self._current_time = dt

        chunk_size = self._index_to_chunk_size[loader_index]
        if chunk_index + 1 == chunk_size:
            self._add_chunk_to_heap(loader_index)

        return loader.symbol, data_type, event

    @_SW.watch("DataQueue")
    def _add_chunk_to_heap(self, loader_index: int):
        chunk = self._next_chunk(loader_index)
        self._index_to_chunk_size[loader_index] = len(chunk)
        for chunk_index, event in enumerate(chunk):
            dt = event.time  # type: ignore
            heapq.heappush(self._event_heap, (dt, loader_index, chunk_index, event))

    @_SW.watch("DataQueue")
    def _next_chunk(self, index: int) -> list[Any]:
        if index not in self._index_to_iterator:
            self._index_to_iterator[index] = self._index_to_loader[index].load(pd.Timestamp(self._current_time, unit="ns"), self._stop)  # type: ignore
        iterator = self._index_to_iterator[index]
        try:
            return next(iterator)
        except StopIteration:
            return []


class SimulatedDataQueueWithThreads(SimulatedDataQueue):
    _loaders: dict[str, list[DataLoader]]

    def __init__(self, workers: int = 4, prefetch_chunk_count: int = 1):
        self._loaders = defaultdict(list)
        self._start = None
        self._stop = None
        self._current_time = None
        self._index_to_loader: dict[int, DataLoader] = {}
        self._index_to_prefetch: dict[int, list[Future]] = defaultdict(list)
        self._index_to_done: dict[int, bool] = defaultdict(bool)
        self._loader_to_index = {}
        self._index_to_chunk_size = {}
        self._index_to_iterator = {}
        self._latest_loader_index = -1
        self._removed_loader_indices = set()
        # TODO: potentially use ProcessPoolExecutor for better performance
        self._pool = ThreadPoolExecutor(max_workers=workers)
        self._prefetch_chunk_count = prefetch_chunk_count

    @property
    def is_running(self) -> bool:
        return self._current_time is not None

    def __add__(self, loader: DataLoader) -> "SimulatedDataQueueWithThreads":
        self._latest_loader_index += 1
        new_loader_index = self._latest_loader_index
        self._loaders[loader.symbol].append(loader)
        self._index_to_loader[new_loader_index] = loader
        self._loader_to_index[loader] = new_loader_index
        if self.is_running:
            self._submit_chunk(new_loader_index)
            self._add_chunk_to_heap(new_loader_index)
        return self

    def __sub__(self, loader: DataLoader) -> "SimulatedDataQueueWithThreads":
        loader_index = self._loader_to_index[loader]
        self._loaders[loader.symbol].remove(loader)
        del self._index_to_loader[loader_index]
        del self._loader_to_index[loader]
        del self._index_to_chunk_size[loader_index]
        del self._index_to_iterator[loader_index]
        del self._index_to_done[loader_index]
        for future in self._index_to_prefetch[loader_index]:
            future.cancel()
        del self._index_to_prefetch[loader_index]
        self._removed_loader_indices.add(loader_index)
        return self

    def get_loader(self, symbol: str, data_type: str) -> DataLoader:
        loaders = self._loaders[symbol]
        for loader in loaders:
            if loader.data_type == data_type:
                return loader
        raise ValueError(f"Loader for {symbol} and {data_type} not found")

    def create_iterable(self, start: str | pd.Timestamp, stop: str | pd.Timestamp) -> Iterator:
        self._start = start
        self._stop = stop
        self._current_time = None
        return self

    def __iter__(self) -> Iterator:
        logger.debug("Initializing chunks for each loader")
        self._current_time = self._start
        self._index_to_chunk_size = {}
        self._index_to_iterator = {}
        self._event_heap = []
        self._submit_chunk_prefetchers()
        for loader_index in self._index_to_loader.keys():
            self._add_chunk_to_heap(loader_index)
        return self

    @_SW.watch("DataQueue")
    def __next__(self) -> tuple[str, str, Any]:
        self._submit_chunk_prefetchers()

        if not self._event_heap:
            raise StopIteration

        loader_index = None

        # get the next event from the heap
        # if the loader_index is in the removed_loader_indices, skip it (optimization to avoid unnecessary heap operations)
        while self._event_heap and (loader_index is None or loader_index in self._removed_loader_indices):
            dt, loader_index, chunk_index, event = heapq.heappop(self._event_heap)

        if loader_index is None or loader_index in self._removed_loader_indices:
            raise StopIteration

        self._current_time = dt
        chunk_size = self._index_to_chunk_size[loader_index]
        if chunk_index + 1 == chunk_size:
            self._add_chunk_to_heap(loader_index)

        loader = self._index_to_loader[loader_index]
        return loader.symbol, loader.data_type, event

    @_SW.watch("DataQueue")
    def _add_chunk_to_heap(self, loader_index: int):
        futures = self._index_to_prefetch[loader_index]
        if not futures and not self._index_to_done[loader_index]:
            loader = self._index_to_loader[loader_index]
            logger.error(f"Error state: No submitted tasks for loader {loader.symbol} {loader.data_type}")
            raise SimulatorError("No submitted tasks for loader")
        elif self._index_to_done[loader_index]:
            return

        # wait for future to finish if needed
        chunk = futures.pop(0).result()
        self._index_to_chunk_size[loader_index] = len(chunk)
        for chunk_index, event in enumerate(chunk):
            dt = event.time  # type: ignore
            heapq.heappush(self._event_heap, (dt, loader_index, chunk_index, event))

    def _next_chunk(self, index: int) -> list[Any]:
        if index not in self._index_to_iterator:
            self._index_to_iterator[index] = self._index_to_loader[index].load(self._current_time, self._stop)  # type: ignore
        iterator = self._index_to_iterator[index]
        try:
            return next(iterator)
        except StopIteration:
            return []

    def _submit_chunk_prefetchers(self):
        for index in self._index_to_loader.keys():
            if len(self._index_to_prefetch[index]) < self._prefetch_chunk_count:
                self._submit_chunk(index)

    def _submit_chunk(self, loader_index: int) -> None:
        future = self._pool.submit(self._next_chunk, loader_index)
        self._index_to_prefetch[loader_index].append(future)

    def __del__(self):
        self._pool.shutdown()


class EventBatcher:
    _BATCH_SETTINGS = {
        "trade": "1Sec",
        "orderbook": "1Sec",
    }

    def __init__(self, source_iterator: Iterator | Iterable, passthrough: bool = False, **kwargs):
        self.source_iterator = source_iterator
        self._passthrough = passthrough
        self._batch_settings = {**self._BATCH_SETTINGS, **kwargs}
        self._batch_settings = {k: pd.Timedelta(v) for k, v in self._batch_settings.items()}

    def __iter__(self) -> Iterator[tuple[str, str, Any]]:
        if self._passthrough:
            _iter = iter(self.source_iterator) if isinstance(self.source_iterator, Iterable) else self.source_iterator
            yield from _iter
            return

        last_symbol: str = None  # type: ignore
        last_data_type: str = None  # type: ignore
        buffer = []
        for symbol, data_type, event in self.source_iterator:
            time: dt_64 = event.time  # type: ignore

            if data_type not in self._batch_settings:
                if buffer:
                    yield last_symbol, last_data_type, self._batch_event(buffer)
                    buffer = []
                yield symbol, data_type, event
                last_symbol, last_data_type = symbol, data_type
                continue

            if symbol != last_symbol:
                if buffer:
                    yield last_symbol, last_data_type, self._batch_event(buffer)
                last_symbol, last_data_type = symbol, data_type
                buffer = [event]
                continue

            if buffer and data_type != last_data_type:
                yield symbol, last_data_type, buffer
                buffer = [event]
                last_symbol, last_data_type = symbol, data_type
                continue

            last_symbol, last_data_type = symbol, data_type
            buffer.append(event)
            if pd.Timedelta(time - buffer[0].time) >= self._batch_settings[data_type]:
                yield symbol, data_type, self._batch_event(buffer)
                buffer = []
                last_symbol, last_data_type = None, None  # type: ignore

        if buffer:
            yield last_symbol, last_data_type, self._batch_event(buffer)

    @staticmethod
    def _batch_event(buffer: list[Any]) -> Any:
        return BatchEvent(buffer[-1].time, buffer) if len(buffer) > 1 else buffer[0]



import click, sys, yaml, sys, time
from os.path import exists, expanduser
import yaml, configparser, socket

from qubx import lookup, logger, formatter
from qubx.core.context import StrategyContextImpl
from qubx.impl.ccxt_connector import CCXTExchangesConnector
from qubx.impl.ccxt_trading import CCXTTradingConnector
from qubx.core.strategy import StrategyContext
from qubx.utils.misc import add_project_to_system_path, Struct, logo, version
from qubx.core.loggers import LogsWriter


LOGFILE = "logs/"


def class_import(name):
    components = name.split(".")
    clz = components[-1]
    mod = __import__(".".join(components[:-1]), fromlist=[clz])
    mod = getattr(mod, clz)
    return mod


def _instruments_for_exchange(exch: str, symbols: list) -> list:
    instrs = []
    for s in symbols:
        instr = lookup.find_symbol(exch.upper(), s.upper())
        if instr is not None:
            instrs.append(instr)
        else:
            logger.warning(f"Can't find instrument for symbol {s} - try to refresh lookup first !")
    return instrs


def load_strategy_config(filename: str) -> Struct:
    with open(filename, "r") as f:
        content = yaml.safe_load(f)

    config = content["config"]
    strat = config["strategy"]
    name = strat.split(".")[-1]
    r = Struct(
        strategy=strat,
        name=name,
        parameters=config.get("parameters", dict()),
        connector=config["connector"],
        exchange=config["exchange"],
        account=config.get("account"),
        md_subscr=config["subscription"],
        strategy_trigger=config["trigger"],
        strategy_fit_trigger=config.get("fit", ""),
        portfolio_logger=config.get("logger", None),
        log_positions_interval=config.get("log_positions_interval", None),
        log_portfolio_interval=config.get("log_portfolio_interval", None),
    )

    universe = config["universe"]
    if isinstance(universe, dict):
        r.instruments = []
        for e, symbs in universe.items():
            r.instruments.extend(_instruments_for_exchange(e, symbs))
    else:
        r.instruments = _instruments_for_exchange(r.exchange, universe)

    return r


def get_account_config(account_id: str, accounts_cfg_file: str) -> dict | None:
    parser = configparser.ConfigParser()
    try:
        parser.optionxform = str  # type: ignore
        parser.read(accounts_cfg_file)
    except Exception as exc:
        logger.error(f"Can't find { accounts_cfg_file } file for reading {account_id} account info: {str(exc)}")
        return None

    if account_id not in parser:
        logger.error(f"No records for {account_id} found in {accounts_cfg_file} file")
        return None

    cfg = dict(parser[account_id])

    # - check if there any reserved funds
    reserves = {}
    if "reserves" in cfg:
        rs = cfg["reserves"]
        for r in rs.split(","):
            s, v = r.strip().split(":")
            reserves[s] = float(v)

    # - add account id and reserves
    return cfg | {"account_id": account_id, "reserves": reserves}


def create_strategy_context(config_file: str, accounts_cfg_file: str, search_paths: list) -> StrategyContext | None:
    cfg = load_strategy_config(config_file)
    try:
        for p in search_paths:
            if exists(pe := expanduser(p)):
                add_project_to_system_path(pe)
        strategy = class_import(cfg.strategy)
    except Exception as err:
        logger.error(str(err))
        return None

    logger.add(LOGFILE + cfg.name + "_{time}.log", format=formatter, rotation="100 MB", colorize=False)

    # - read account creds
    acc_config = {}
    if cfg.account is not None:
        acc_config = get_account_config(cfg.account, accounts_cfg_file)
        if acc_config is None:
            return None

    # - check connector
    conn = cfg.connector.lower()
    match conn:
        case "ccxt":
            # - TODO: we need some factory here
            broker = CCXTTradingConnector(cfg.exchange.lower(), **acc_config)
            exchange_connector = CCXTExchangesConnector(cfg.exchange.lower(), broker, **acc_config)
        case _:
            raise ValueError(f"Connector {conn} is not supported yet !")

    # - generate new run id
    run_id = socket.gethostname() + "-" + str(broker.time().item() // 100_000_000)

    # - get logger
    writer = None
    _w_class = cfg.portfolio_logger
    if _w_class is not None:
        if "." not in _w_class:
            _w_class = "qubx.core.loggers." + _w_class
        try:
            w_class = class_import(_w_class)
            writer = w_class(acc_config["account_id"], strategy.__name__, run_id)
        except Exception as err:
            logger.warning(f"Can't instantiate specified writer {_w_class}: {str(err)}")
            writer = LogsWriter(acc_config["account_id"], strategy.__name__, run_id)

    logger.info(
        f""" - - - <blue>Qubx</blue> (ver. <red>{version()}</red>) - - -\n - Strategy: {strategy}\n - Config: {cfg.parameters} """
    )
    ctx = StrategyContextImpl(
        strategy,
        cfg.parameters,
        exchange_connector,
        instruments=cfg.instruments,
        md_subscription=cfg.md_subscr,
        trigger_spec=cfg.strategy_trigger,
        fit_spec=cfg.strategy_fit_trigger,
        logs_writer=writer,
        positions_log_freq=cfg.log_positions_interval,
        portfolio_log_freq=cfg.log_portfolio_interval,
    )

    return ctx


def _run_in_jupyter(filename: str, accounts: str, paths: list):
    """
    Helper for run this in jupyter console
    """
    try:
        from jupyter_console.app import ZMQTerminalIPythonApp
    except ImportError:
        logger.error("Can't find <red>ZMQTerminalIPythonApp</red> module - try to install jupyter package first")
        return
    try:
        import nest_asyncio
    except ImportError:
        logger.error("Can't find <red>nest_asyncio</red> module - try to install it first")
        return

    class TerminalRunner(ZMQTerminalIPythonApp):
        def __init__(self, **kwargs) -> None:
            self.init_code = kwargs.pop("init_code")
            super().__init__(**kwargs)

        def init_banner(self):
            pass

        def initialize(self, argv=None):
            super().initialize(argv=[])
            self.shell.run_cell(self.init_code)

    logger.info("Running in Jupyter console")
    TerminalRunner.launch_instance(
        init_code=f"""
# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
import qubx
%qubxd
import pandas as pd
import nest_asyncio; nest_asyncio.apply()
from qubx.utils.misc import dequotify, quotify
from qubx.utils.runner import create_strategy_context
from qubx.pandaz.utils import *
import qubx.pandaz.ta as pta

# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
ctx = create_strategy_context('{filename}', '{accounts}', {paths})
ctx.start()

# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
def orders(symbol=None):
    return ctx.exchange_service.get_orders(symbol)

def trade(symbol, qty, price=None, tif='gtc'):
    return ctx.trade(symbol, qty, price, tif)

# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
def pnl_report(all=True):
    from tabulate import tabulate
    d = dict()
    for s, p in ctx.positions.items():
        mv = round(p.market_value_funds, 3)
        if mv != 0.0 or all:
            d[dequotify(s)] = dict(
                Position=round(p.quantity, p.instrument.size_precision),  
                PnL=p.total_pnl(), 
                AvgPrice=round(p.position_avg_price_funds, p.instrument.price_precision), 
                LastPrice=round(p.last_update_price, p.instrument.price_precision),
                MktValue=mv
            )
    d = pd.DataFrame.from_dict(d).T
    # d = d[d['PnL'] != 0.0]
    if d.empty:
        print('-(no open positions yet)-')
        return
    d = d.sort_values('PnL' ,ascending=False)
    # d = pd.concat((d, pd.Series(dict(TOTAL=d['PnL'].sum()), name='PnL'))).fillna('')
    d = pd.concat((d, scols(pd.Series(dict(TOTAL=d['PnL'].sum()), name='PnL'), pd.Series(dict(TOTAL=d['MktValue'].sum()), name='MktValue')))).fillna('')
    print(tabulate(d, ['Position', 'PnL', 'AvgPrice', 'LastPrice', 'MktValue'], tablefmt='rounded_grid'))
# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
__exit = exit
def exit():
    ctx.stop(); __exit()
# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
"""
    )


@click.command()
@click.argument("filename", type=click.Path(exists=True))
@click.option("--accounts", "-a", default="accounts.cfg", type=click.STRING, help="Live accounts configuration file")
@click.option(
    "--paths",
    "-p",
    multiple=True,
    default=["../", "~/projects/"],
    type=click.STRING,
    help="Live accounts configuration file",
)
@click.option("--jupyter", "-j", is_flag=True, default=False, help="Run strategy in jupyter console", show_default=True)
def run(filename: str, accounts: str, paths: list, jupyter: bool):
    if jupyter:
        _run_in_jupyter(filename, accounts, paths)
        return

    # - show Qubx logo with current version
    logo()

    # - create context
    ctx = create_strategy_context(filename, accounts, paths)
    if ctx is None:
        return

    # - run main loop
    try:
        ctx.start()

        # - just wake up every 60 sec and check if it's OK
        while True:
            time.sleep(60)

    except KeyboardInterrupt:
        ctx.stop()
        time.sleep(1)
        sys.exit(0)


if __name__ == "__main__":
    run()



from threading import Thread
import threading
from typing import Any, Dict, List, Optional

import asyncio
from asyncio.tasks import Task
from asyncio.events import AbstractEventLoop
from collections import defaultdict

import ccxt.pro as cxp
from ccxt.base.exchange import Exchange
from ccxt import NetworkError, ExchangeClosedByUser

import re
import numpy as np
import pandas as pd

from qubx import logger
from qubx.core.basics import Instrument, Position, dt_64, Deal, CtrlChannel
from qubx.core.helpers import BasicScheduler
from qubx.core.strategy import IBrokerServiceProvider, ITradingServiceProvider
from qubx.core.series import TimeSeries, Bar, Trade, Quote
from qubx.impl.ccxt_utils import DATA_PROVIDERS_ALIASES, ccxt_convert_trade
from qubx.utils.ntp import time_now

# - register custom wrappers
from .ccxt_customizations import BinanceQV, BinanceQVUSDM

cxp.binanceqv = BinanceQV  # type: ignore
cxp.binanceqv_usdm = BinanceQVUSDM  # type: ignore
cxp.exchanges.append("binanceqv")
cxp.exchanges.append("binanceqv_usdm")


class CCXTExchangesConnector(IBrokerServiceProvider):
    _exchange: Exchange
    _subsriptions: Dict[str, List[str]]
    _scheduler: BasicScheduler | None = None

    _last_quotes: Dict[str, Optional[Quote]]
    _loop: AbstractEventLoop
    _thread_event_loop: Thread

    def __init__(
        self,
        exchange_id: str,
        trading_service: ITradingServiceProvider,
        **exchange_auth,
    ):
        super().__init__(exchange_id, trading_service)
        self.trading_service = trading_service
        exchange_id = exchange_id.lower()

        # - setup communication bus
        self.set_communication_channel(bus := CtrlChannel("databus", sentinel=(None, None, None)))
        self.trading_service.set_communication_channel(bus)

        # - init CCXT stuff
        exch = DATA_PROVIDERS_ALIASES.get(exchange_id, exchange_id)
        if exch not in cxp.exchanges:
            raise ValueError(f"Exchange {exchange_id} -> {exch} is not supported by CCXT.pro !")

        # - create new even loop
        self._loop = asyncio.new_event_loop()
        asyncio.set_event_loop(self._loop)

        # - create exchange's instance
        self._exchange = getattr(cxp, exch)(exchange_auth | {"asyncio_loop": self._loop})
        self._last_quotes = defaultdict(lambda: None)
        self._subsriptions = defaultdict(list)

        logger.info(f"{exchange_id} initialized - current time {self.trading_service.time()}")

    def get_scheduler(self) -> BasicScheduler:
        # - standard scheduler
        if self._scheduler is None:
            self._scheduler = BasicScheduler(self.get_communication_channel(), lambda: self.time().item())
        return self._scheduler

    def subscribe(
        self, subscription_type: str, instruments: List[Instrument], timeframe: Optional[str] = None, nback: int = 0
    ) -> bool:
        to_process = self._check_existing_subscription(subscription_type.lower(), instruments)
        if not to_process:
            logger.info(f"Symbols {to_process} already subscribed on {subscription_type} data")
            return False

        # - subscribe to market data updates
        match sbscr := subscription_type.lower():
            case "ohlc":
                if timeframe is None:
                    raise ValueError("timeframe must not be None for OHLC data subscription")

                # convert to exchange format
                tframe = self._get_exch_timeframe(timeframe)
                for s in to_process:
                    # self._task_a(self._listen_to_ohlcv(self.get_communication_channel(), s, tframe, nback))
                    asyncio.run_coroutine_threadsafe(
                        self._listen_to_ohlcv(self.get_communication_channel(), s, tframe, nback), self._loop
                    )
                    self._subsriptions[sbscr].append(s.lower())
                logger.info(f"Subscribed on {sbscr} updates for {len(to_process)} symbols: \n\t\t{to_process}")

            case "trades":
                if timeframe is None:
                    raise ValueError("timeframe must not be None for trade data subscription")
                tframe = self._get_exch_timeframe(timeframe)
                for s in to_process:
                    asyncio.run_coroutine_threadsafe(
                        self._listen_to_trades(self.get_communication_channel(), s, tframe, nback), self._loop
                    )
                    self._subsriptions[sbscr].append(s.lower())
                logger.info(f"Subscribed on {sbscr} updates for {len(to_process)} symbols: \n\t\t{to_process}")

            case "quotes":
                raise ValueError("TODO")

            case _:
                raise ValueError("TODO")

        # - subscibe to executions reports
        for s in to_process:
            asyncio.run_coroutine_threadsafe(
                self._listen_to_execution_reports(self.get_communication_channel(), s), self._loop
            )

        self._thread_event_loop = Thread(target=self._loop.run_forever, args=(), daemon=True)
        self._thread_event_loop.start()

        return True

    def _check_existing_subscription(self, subscription_type, instruments: List[Instrument]) -> List[str]:
        subscribed = self._subsriptions[subscription_type]
        to_subscribe = []
        for s in instruments:
            if s.symbol not in subscribed:
                to_subscribe.append(s.symbol)
        return to_subscribe

    def _time_msec_nbars_back(self, timeframe: str, nbarsback: int) -> int:
        return (self.time() - nbarsback * pd.Timedelta(timeframe)).asm8.item() // 1000000

    def get_historical_ohlcs(self, symbol: str, timeframe: str, nbarsback: int) -> List[Bar]:
        assert nbarsback >= 1
        since = self._time_msec_nbars_back(timeframe, nbarsback)

        # - retrieve OHLC data
        # - TODO: check if nbarsback > max_limit (1000) we need to do more requests
        # - TODO: how to get quoted volumes ?
        async def _get():
            return await self._exchange.fetch_ohlcv(symbol, self._get_exch_timeframe(timeframe), since=since, limit=nbarsback + 1)  # type: ignore

        fut = asyncio.run_coroutine_threadsafe(_get(), self._loop)
        res = fut.result(60)

        _arr = []
        for oh in res:  # type: ignore
            _arr.append(
                Bar(oh[0] * 1_000_000, oh[1], oh[2], oh[3], oh[4], oh[6], oh[7])
                if len(oh) > 6
                else Bar(oh[0] * 1_000_000, oh[1], oh[2], oh[3], oh[4], oh[5])
            )
        return _arr

    async def _listen_to_execution_reports(self, channel: CtrlChannel, symbol: str):
        while channel.control.is_set():
            try:
                exec = await self._exchange.watch_orders(symbol)  # type: ignore
                _msg = f"\nexecs_{symbol} = [\n"
                for report in exec:
                    _msg += "\t" + str(report) + ",\n"
                    order, deals = self.trading_service.process_execution_report(symbol, report)
                    # - send update to client
                    channel.send((symbol, "order", order))
                    if deals:
                        channel.send((symbol, "deals", deals))
                logger.debug(_msg + "]\n")

            except NetworkError as e:
                logger.error(f"(CCXTDataConnector) NetworkError in _listen_to_execution_reports : {e}")
                await asyncio.sleep(1)
                continue

            except ExchangeClosedByUser:
                # - we closed connection so just stop it
                logger.info(f"(CCXTDataConnector) {symbol} execution reports listening has been stopped")
                break

            except Exception as err:
                logger.error(f"(CCXTDataConnector) exception in _listen_to_execution_reports : {err}")
                logger.exception(err)

    async def _listen_to_ohlcv(self, channel: CtrlChannel, symbol: str, timeframe: str, nbarsback: int):
        # - check if we need to load initial 'snapshot'
        if nbarsback >= 1:
            start = self._time_msec_nbars_back(timeframe, nbarsback)
            ohlcv = await self._exchange.fetch_ohlcv(symbol, timeframe, since=start, limit=nbarsback + 1)  # type: ignore
            # - just send data as the list
            channel.send(
                (
                    symbol,
                    "hist_bars",
                    [Bar(oh[0] * 1_000_000, oh[1], oh[2], oh[3], oh[4], oh[6], oh[7]) for oh in ohlcv],
                )
            )
            logger.info(f"{symbol}: loaded {len(ohlcv)} {timeframe} bars")

        while channel.control.is_set():
            try:
                ohlcv = await self._exchange.watch_ohlcv(symbol, timeframe)  # type: ignore

                # - update positions by actual close price
                last_close = ohlcv[-1][4]

                # - there is no single method to get OHLC update's event time for every broker
                # - for instance it's possible to do for Binance but for example Bitmex doesn't provide such info
                # - so we will use ntp adjusted time here
                self.trading_service.update_position_price(symbol, self.time(), last_close)

                for oh in ohlcv:
                    channel.send((symbol, "bar", Bar(oh[0] * 1_000_000, oh[1], oh[2], oh[3], oh[4], oh[6], oh[7])))

            except NetworkError as e:
                logger.error(f"(CCXTDataConnector) NetworkError in _listen_to_ohlcv : {e}")
                await asyncio.sleep(1)
                continue

            except ExchangeClosedByUser:
                # - we closed connection so just stop it
                logger.info(f"(CCXTDataConnector) {symbol} OHLCV listening has been stopped")
                break

            except Exception as e:
                # logger.error(str(e))
                logger.error(f"(CCXTDataConnector) exception in _listen_to_ohlcv : {e}")
                logger.exception(e)
                await self._exchange.close()  # type: ignore
                raise e

    async def _listen_to_trades(self, channel: CtrlChannel, symbol: str, timeframe: str, nbarsback: int):
        """
        Listen to trades for the given symbol. We also load historical ohlcv data in the beginning.
        """
        # - check if we need to load initial 'snapshot'
        logger.info(f"Listening to trades for {symbol} {timeframe} {nbarsback}...")
        if nbarsback >= 1:
            start = self._time_msec_nbars_back(timeframe, nbarsback)
            ohlcv = await self._exchange.fetch_ohlcv(symbol, timeframe, since=start, limit=nbarsback + 1)  # type: ignore
            # - just send data as the list
            channel.send(
                (
                    symbol,
                    "hist_bars",
                    [Bar(oh[0] * 1_000_000, oh[1], oh[2], oh[3], oh[4], oh[6], oh[7]) for oh in ohlcv],
                )
            )
            logger.info(f"{symbol}: loaded {len(ohlcv)} {timeframe} bars")

        while channel.control.is_set():
            try:
                trades = await self._exchange.watch_trades(symbol)

                # - update positions by actual close price
                last_trade = ccxt_convert_trade(trades[-1])
                self.trading_service.update_position_price(symbol, last_trade.time, last_trade)

                for trade in trades:
                    channel.send((symbol, "trade", ccxt_convert_trade(trade)))

            except NetworkError as e:
                logger.error(f"(CCXTDataConnector) NetworkError in _listen_to_trades : {e}")
                await asyncio.sleep(1)
                continue

            except ExchangeClosedByUser:
                # - we closed connection so just stop it
                logger.info(f"(CCXTDataConnector) {symbol} Trades listening has been stopped")
                break

            except Exception as e:
                logger.error(f"(CCXTDataConnector) exception in _listen_to_trades : {e}")
                logger.exception(e)
                await self._exchange.close()  # type: ignore
                raise e

    def get_quote(self, symbol: str) -> Optional[Quote]:
        return self._last_quotes[symbol]

    def _get_exch_timeframe(self, timeframe: str) -> str:
        if timeframe is not None:
            _t = re.match(r"(\d+)(\w+)", timeframe)
            timeframe = f"{_t[1]}{_t[2][0].lower()}" if _t and len(_t.groups()) > 1 else timeframe

        tframe = self._exchange.find_timeframe(timeframe)
        if tframe is None:
            raise ValueError(f"timeframe {timeframe} is not supported by {self._exchange.name}")

        return tframe

    def close(self):
        try:
            # self._loop.run_until_complete(self._exchange.close())  # type: ignore
            asyncio.run_coroutine_threadsafe(self._exchange.close(), self._loop)
        except Exception as e:
            logger.error(e)

    def time(self) -> dt_64:
        """
        Returns current time as dt64
        """
        return time_now()



"""
 # All interfaces related to strategy etc
"""

from typing import Any, Callable, Dict, List, Optional, Tuple, Union
import traceback

import pandas as pd

from qubx import lookup, logger
from qubx.core.account import AccountProcessor
from qubx.core.helpers import BasicScheduler, set_parameters_to_object
from qubx.core.basics import (
    TargetPosition,
    TriggerEvent,
    Deal,
    Instrument,
    Order,
    Position,
    Signal,
    dt_64,
    td_64,
    ITimeProvider,
    IComminucationManager,
)
from qubx.core.series import Trade, Quote, Bar, OHLCV
from qubx.utils.misc import Stopwatch


_SW = Stopwatch()


class ITradingServiceProvider(ITimeProvider, IComminucationManager):
    """
    Trading service provider interface that manages account operations, order placement, and position tracking.
    """

    acc: AccountProcessor

    def set_account(self, account: AccountProcessor):
        """
        Set the account processor for the trading service provider.

        :param account: The AccountProcessor object to be set.
        """
        self.acc = account

    def get_account(self) -> AccountProcessor:
        """
        Retrieve the current account processor.

        :return: The current AccountProcessor object.
        """
        return self.acc

    def get_name(self) -> str:
        """
        Get the name of the trading service provider.

        :return: The name of the trading service provider as a string.
        :raises NotImplementedError: If the method is not implemented by the subclass.
        """
        raise NotImplementedError("get_name is not implemented")

    def get_account_id(self) -> str:
        """
        Get the account ID associated with the trading service provider.

        :return: The account ID as a string.
        :raises NotImplementedError: If the method is not implemented by the subclass.
        """
        raise NotImplementedError("get_account_id is not implemented")

    def get_capital(self) -> float:
        """
        Get the available capital in the account.

        :return: The free capital as a float.
        """
        return self.acc.get_free_capital()

    def send_order(
        self,
        instrument: Instrument,
        order_side: str,
        order_type: str,
        amount: float,
        price: float | None = None,
        client_id: str | None = None,
        time_in_force: str = "gtc",
        **optional,
    ) -> Order:
        """
        Send an order to the trading service.

        :param instrument: The instrument to trade.
        :param order_side: The side of the order (e.g., "buy" or "sell").
        :param order_type: The type of the order (e.g., "market" or "limit").
        :param amount: The amount of the instrument to trade.
        :param price: The price for limit orders (optional).
        :param client_id: A client-specified ID for the order (optional).
        :param time_in_force: The time in force for the order (default is "gtc" - good till cancelled).
        :param optional: Additional optional parameters for the order.
        :return: An Order object representing the sent order.
        :raises NotImplementedError: If the method is not implemented by the subclass.
        """
        raise NotImplementedError("send_order is not implemented")

    def cancel_order(self, order_id: str) -> Order | None:
        """
        Cancel an existing order.

        :param order_id: The ID of the order to cancel.
        :return: The cancelled Order object if successful, None otherwise.
        :raises NotImplementedError: If the method is not implemented by the subclass.
        """
        raise NotImplementedError("cancel_order is not implemented")

    def get_orders(self, symbol: str | None = None) -> List[Order]:
        """
        Get a list of current orders, optionally filtered by symbol.

        :param symbol: The symbol to filter orders by (optional).
        :return: A list of Order objects.
        :raises NotImplementedError: If the method is not implemented by the subclass.
        """
        raise NotImplementedError("get_orders is not implemented")

    def get_position(self, instrument: Instrument | str) -> Position:
        """
        Get the current position for a given instrument.

        :param instrument: The instrument or symbol to get the position for.
        :return: A Position object representing the current position.
        :raises NotImplementedError: If the method is not implemented by the subclass.
        """
        raise NotImplementedError("get_position is not implemented")

    def get_base_currency(self) -> str:
        """
        Get the base currency for the account.

        :return: The base currency as a string.
        :raises NotImplementedError: If the method is not implemented by the subclass.
        """
        raise NotImplementedError("get_basic_currency is not implemented")

    def process_execution_report(self, symbol: str, report: Dict[str, Any]) -> Tuple[Order, List[Deal]]:
        """
        Process an execution report for a given symbol.

        :param symbol: The symbol the execution report is for.
        :param report: A dictionary containing the execution report details.
        :return: A tuple containing the updated Order and a list of Deal objects.
        :raises NotImplementedError: If the method is not implemented by the subclass.
        """
        raise NotImplementedError("process_execution_report is not implemented")

    @staticmethod
    def _extract_price(update: float | Quote | Trade | Bar) -> float:
        """
        Extract the price from various types of market data updates.

        :param update: The market data update, which can be a float, Quote, Trade, or Bar.
        :return: The extracted price as a float.
        :raises ValueError: If the update type is unknown.
        """
        if isinstance(update, float):
            return update
        elif isinstance(update, Quote):
            return 0.5 * (update.bid + update.ask)  # type: ignore
        elif isinstance(update, Trade):
            return update.price  # type: ignore
        elif isinstance(update, Bar):
            return update.close  # type: ignore
        else:
            raise ValueError(f"Unknown update type: {type(update)}")

    def update_position_price(self, symbol: str, timestamp: dt_64, update: float | Quote | Trade | Bar):
        """
        Update the price of a position for a given symbol.

        :param symbol: The symbol of the position to update.
        :param timestamp: The timestamp of the update.
        :param update: The price update, which can be a float, Quote, Trade, or Bar.
        """
        self.acc.update_position_price(timestamp, symbol, ITradingServiceProvider._extract_price(update))


class IBrokerServiceProvider(IComminucationManager, ITimeProvider):
    trading_service: ITradingServiceProvider

    def __init__(self, exchange_id: str, trading_service: ITradingServiceProvider) -> None:
        self._exchange_id = exchange_id
        self.trading_service = trading_service

    def subscribe(self, subscription_type: str, instruments: List[Instrument], **kwargs) -> bool:
        raise NotImplementedError("subscribe")

    def unsubscribe(self, subscription_type: str, instruments: List[Instrument]) -> bool:
        raise NotImplementedError("unsubscribe")

    def has_subscription(self, subscription_type: str, instrument: Instrument) -> bool:
        raise NotImplementedError("has_subscription")

    def get_historical_ohlcs(self, symbol: str, timeframe: str, nbarsback: int) -> List[Bar]:
        raise NotImplementedError("get_historical_ohlcs")

    def get_quote(self, symbol: str) -> Quote | None:
        raise NotImplementedError("get_quote")

    def get_trading_service(self) -> ITradingServiceProvider:
        return self.trading_service

    def close(self):
        pass

    def get_scheduler(self) -> BasicScheduler:
        raise NotImplementedError("schedule_event")

    @property
    def is_simulated_trading(self) -> bool:
        return False


class SubscriptionType:
    """
    Subscription type constants
    """

    QUOTE = "quote"
    TRADE = "trade"
    AGG_TRADE = "agg_trade"
    OHLC = "ohlc"


class IMarketDataProvider(ITimeProvider):
    """
    Interface for market data providing class
    """

    def ohlc(self, instrument: Instrument | str, timeframe: str) -> OHLCV: ...

    def quote(self, symbol: str) -> Quote | None: ...

    def get_historical_ohlcs(self, instrument: Instrument | str, timeframe: str, length: int) -> OHLCV | None: ...

    def get_aux_data(self, data_id: str, **parametes) -> pd.DataFrame | None: ...

    def get_instruments(self) -> List[Instrument]: ...

    def get_instrument(self, symbol: str) -> Instrument | None: ...


class StrategyContext(IMarketDataProvider):
    """
    Strategy context interface
    """

    instruments: List[Instrument]  # list of instruments this strategy trades
    positions: Dict[str, Position]  # positions of the strategy (instrument -> position)
    acc: AccountProcessor
    broker_provider: IBrokerServiceProvider  # market data provider

    def process_data(self, symbol: str, d_type: str, data: Any) -> bool: ...

    def start(self, blocking: bool = False): ...

    def stop(self): ...

    def trade(
        self,
        instr_or_symbol: Instrument | str,
        amount: float,
        price: float | None = None,
        time_in_force="gtc",
        **options,
    ) -> Order: ...

    def cancel(self, instr_or_symbol: Instrument | str): ...

    def cancel_order(self, order_id: str): ...

    def get_capital(self) -> float: ...

    def get_reserved(self, instrument: Instrument) -> float: ...

    def set_universe(self, instruments: list[Instrument]): ...

    def subscribe(self, subscription_type: str, instr_or_symbol: Instrument | str, **kwargs) -> bool: ...

    def unsubscribe(self, subscription_type: str, instr_or_symbol: Instrument | str) -> bool: ...

    def has_subscription(self, subscription_type: str, instr_or_symbol: Instrument | str) -> bool: ...

    @staticmethod
    def get_latencies_report():
        scope_to_latency_sec = {scope: _SW.latency_sec(scope) for scope in _SW.latencies.keys()}
        scope_to_count = {l: _SW.counts[l] for l in scope_to_latency_sec.keys()}
        scope_to_total_time = {scope: scope_to_count[scope] * lat for scope, lat in scope_to_latency_sec.items()}
        # create pandas datafrmae from dictionaries
        lats = pd.DataFrame(
            {
                "scope": list(scope_to_latency_sec.keys()),
                "latency": list(scope_to_latency_sec.values()),
                "count": list(scope_to_count.values()),
                "total_time": list(scope_to_total_time.values()),
            }
        )
        lats["latency"] = lats["latency"].apply(lambda x: f"{x:.4f}")
        lats["total_time (min)"] = lats["total_time"].apply(lambda x: f"{x / 60:.4f}")
        lats.drop(columns=["total_time"], inplace=True)
        return lats


class IPositionGathering:
    """
    Common interface for position gathering
    """

    def alter_position_size(self, ctx: StrategyContext, target: TargetPosition) -> float: ...

    def alter_positions(
        self, ctx: StrategyContext, targets: List[TargetPosition] | TargetPosition
    ) -> Dict[Instrument, float]:
        if not isinstance(targets, list):
            targets = [targets]

        res = {}
        if targets:
            for t in targets:
                if t.is_service:  # we skip processing service positions
                    continue
                try:
                    res[t.instrument] = self.alter_position_size(ctx, t)
                except Exception as ex:
                    logger.error(f"[{ctx.time()}]: Failed processing target position {t} : {ex}")
                    logger.opt(colors=False).error(traceback.format_exc())
        return res

    def on_execution_report(self, ctx: StrategyContext, instrument: Instrument, deal: Deal): ...


class IPositionSizer:
    """
    Common interface for get actual positions from signals
    """

    def calculate_target_positions(self, ctx: StrategyContext, signals: List[Signal]) -> List[TargetPosition]:
        """
        Target position size calculator

        :param ctx: strategy context object
        :param signals: list of signals to process
        """
        raise NotImplementedError("calculate_target_positions is not implemented")


class PositionsTracker:
    """
    Process signals from strategy and track position. It can contains logic for risk management for example.
    """

    _sizer: IPositionSizer

    def __init__(self, sizer: IPositionSizer) -> None:
        self._sizer = sizer

    def get_position_sizer(self) -> IPositionSizer:
        return self._sizer

    def is_active(self, instrument: Instrument) -> bool:
        return True

    def process_signals(self, ctx: StrategyContext, signals: List[Signal]) -> List[TargetPosition] | TargetPosition:
        """
        Default implementation just returns calculated target positions
        """
        return self.get_position_sizer().calculate_target_positions(ctx, signals)

    def update(
        self, ctx: StrategyContext, instrument: Instrument, update: Quote | Trade | Bar
    ) -> List[TargetPosition] | TargetPosition:
        """
        Tracker is being updated by new market data.
        It may require to change position size or create new position because of interior tracker's logic (risk management for example).
        """
        ...

    def on_execution_report(self, ctx: StrategyContext, instrument: Instrument, deal: Deal):
        """
        Tracker is notified when execution report is received
        """
        ...


class IStrategy:
    ctx: StrategyContext

    def __init__(self, **kwargs) -> None:
        set_parameters_to_object(self, **kwargs)

    def on_start(self, ctx: StrategyContext):
        """
        This method is called strategy is started
        """
        pass

    def on_fit(
        self, ctx: StrategyContext, fit_time: str | pd.Timestamp, previous_fit_time: str | pd.Timestamp | None = None
    ):
        """
        This method is called when it's time to fit model
        :param fit_time: last time of fit data to use
        :param previous_fit_time: last time of fit data used in previous fit call
        """
        return None

    def on_universe_change(
        self, ctx: StrategyContext, add_instruments: list[Instrument], rm_instruments: list[Instrument]
    ) -> None:
        """
        This method is called when the trading universe is updated.
        """
        return None

    def on_event(self, ctx: StrategyContext, event: TriggerEvent) -> List[Signal] | Signal | None:
        return None

    def on_stop(self, ctx: StrategyContext):
        pass

    def tracker(self, ctx: StrategyContext) -> PositionsTracker | None:
        pass



from typing import Any, Dict, List, Optional, Tuple
import pandas as pd

from qubx import logger
from qubx.core.basics import Order, Deal, Position
from qubx.core.series import TimeSeries, Bar, Trade, Quote


EXCHANGE_ALIASES = {"binance.um": "binanceusdm", "binance.cm": "binancecoinm", "kraken.f": "krakenfutures"}

DATA_PROVIDERS_ALIASES = EXCHANGE_ALIASES | {"binance": "binanceqv", "binance.um": "binanceqv_usdm"}


def ccxt_convert_order_info(symbol: str, raw: Dict[str, Any]) -> Order:
    """
    Convert CCXT excution record to Order object
    """
    ri = raw["info"]
    amnt = float(ri.get("origQty", raw.get("amount")))
    price = raw["price"]
    status = raw["status"]
    side = raw["side"].upper()
    _type = ri.get("type", raw.get("type")).upper()
    if status == "open":
        status = ri.get("status", status)  # for filled / part_filled ?

    return Order(
        id=raw["id"],
        type=_type,
        symbol=symbol,
        time=pd.Timestamp(raw["timestamp"], unit="ms"),  # type: ignore
        quantity=amnt,
        price=float(price) if price is not None else 0.0,
        side=side,
        status=status.upper(),
        time_in_force=raw["timeInForce"],
        client_id=raw["clientOrderId"],
        cost=float(raw["cost"] or 0),  # cost can be None
    )


def ccxt_convert_deal_info(raw: Dict[str, Any]) -> Deal:
    fee_amount = None
    fee_currency = None
    if "fee" in raw:
        fee_amount = float(raw["fee"]["cost"])
        fee_currency = raw["fee"]["currency"]
    return Deal(
        id=raw["id"],
        order_id=raw["order"],
        time=pd.Timestamp(raw["timestamp"], unit="ms"),  # type: ignore
        amount=float(raw["amount"]) * (-1 if raw["side"] == "sell" else +1),
        price=float(raw["price"]),
        aggressive=raw["takerOrMaker"] == "taker",
        fee_amount=fee_amount,
        fee_currency=fee_currency,
    )


def ccxt_extract_deals_from_exec(report: Dict[str, Any]) -> List[Deal]:
    """
    Small helper for extracting deals (trades) from CCXT execution report
    """
    deals = list()
    if trades := report.get("trades"):
        for t in trades:
            deals.append(ccxt_convert_deal_info(t))
    return deals


def ccxt_restore_position_from_deals(
    pos: Position, current_volume: float, deals: List[Deal], reserved_amount: float = 0.0
) -> Position:
    if current_volume != 0:
        instr = pos.instrument
        _last_deals = []

        # - try to find last deals that led to this position
        for d in sorted(deals, key=lambda x: x.time, reverse=True):
            current_volume -= d.amount
            # - spot case when fees may be deducted from the base coin
            #   that may decrease total amount
            if d.fee_amount is not None:
                if instr.base == d.fee_currency:
                    current_volume += d.fee_amount
            # print(d.amount, current_volume)
            _last_deals.insert(0, d)

            # - take in account reserves
            if abs(current_volume) - abs(reserved_amount) < instr.min_size_step:
                break

        # - reset to 0
        pos.reset()

        if abs(current_volume) - abs(reserved_amount) > instr.min_size_step:
            # - - - TODO - - - !!!!
            logger.warning(
                f"Couldn't restore full deals history for {instr.symbol} symbol. Qubx will use zero position !"
            )
        else:
            fees_in_base = 0.0
            for d in _last_deals:
                pos.update_position_by_deal(d)
                if d.fee_amount is not None:
                    if instr.base == d.fee_currency:
                        fees_in_base += d.fee_amount
            # - we round fees up in case of fees in base currency
            pos.quantity -= pos.instrument.round_size_up(fees_in_base)
    return pos


def ccxt_convert_trade(trade: dict[str, Any]) -> Trade:
    t_ns = trade["timestamp"] * 1_000_000  # this is trade time
    s, info, price, amnt = trade["symbol"], trade["info"], trade["price"], trade["amount"]
    m = info["m"]
    return Trade(t_ns, price, amnt, int(not m), int(trade["id"]))



from typing import List
import numpy as np

from qubx import logger
from qubx.core.basics import Position, Signal, TargetPosition
from qubx.core.strategy import IPositionSizer, StrategyContext


class FixedSizer(IPositionSizer):
    """
    Simplest fixed sizer class. It uses same fixed size for all signals.
    We use it for quick backtesting of generated signals in most cases.
    """

    def __init__(self, fixed_size: float, amount_in_quote: bool = True):
        self.amount_in_quote = amount_in_quote
        self.fixed_size = abs(fixed_size)

    def calculate_target_positions(self, ctx: StrategyContext, signals: List[Signal]) -> List[TargetPosition]:
        if not self.amount_in_quote:
            return [TargetPosition.create(ctx, s, s.signal * self.fixed_size) for s in signals]
        positions = []
        for signal in signals:
            q = ctx.quote(signal.instrument.symbol)
            if q is None:
                logger.error(
                    f"{self.__class__.__name__}: Can't get actual market quote for {signal.instrument.symbol} !"
                )
                continue
            positions.append(TargetPosition.create(ctx, signal, signal.signal * self.fixed_size / q.mid_price()))
        return positions


class FixedLeverageSizer(IPositionSizer):
    """
    Defines the leverage per each unit of signal. If leverage is 1.0, then
    the position leverage will be equal to the signal value.
    """

    def __init__(self, leverage: float):
        """
        Args:
            leverage (float): leverage value per a unit of signal.
            split_by_symbols (bool): Should the calculated leverage by divided
            by the number of symbols in the universe.
        """
        self.leverage = leverage

    def calculate_target_positions(self, ctx: StrategyContext, signals: List[Signal]) -> List[TargetPosition]:
        total_capital = ctx.acc.get_total_capital()
        positions = []
        for signal in signals:
            q = ctx.quote(signal.instrument.symbol)
            if q is None:
                logger.error(
                    f"{self.__class__.__name__}: Can't get actual market quote for {signal.instrument.symbol} !"
                )
                continue
            size = signal.signal * self.leverage * total_capital / q.mid_price() / len(ctx.instruments)
            positions.append(TargetPosition.create(ctx, signal, size))
        return positions


class FixedRiskSizer(IPositionSizer):
    def __init__(
        self,
        max_cap_in_risk: float,
        max_allowed_position=np.inf,
        reinvest_profit: bool = True,
        divide_by_symbols: bool = True,
    ):
        """
        Create fixed risk sizer calculator instance.
        :param max_cap_in_risk: maximal risked capital (in percentage)
        :param max_allowed_position: limitation for max position size in quoted currency (i.e. max 5000 in USDT)
        :param reinvest_profit: if true use profit to reinvest
        """
        self.max_cap_in_risk = max_cap_in_risk / 100
        self.max_allowed_position_quoted = max_allowed_position
        self.reinvest_profit = reinvest_profit
        self.divide_by_symbols = divide_by_symbols

    def calculate_target_positions(self, ctx: StrategyContext, signals: List[Signal]) -> List[TargetPosition]:
        t_pos = []
        for signal in signals:
            target_position_size = 0
            if signal.signal != 0:
                if signal.stop and signal.stop > 0:
                    _pos = ctx.positions[signal.instrument.symbol]
                    _q = ctx.quote(signal.instrument.symbol)

                    _direction = np.sign(signal.signal)
                    # - hey, we can't trade using negative balance ;)
                    _cap = max(ctx.acc.get_total_capital() if self.reinvest_profit else ctx.acc.get_free_capital(), 0)
                    _entry = _q.ask if _direction > 0 else _q.bid
                    # fmt: off
                    target_position_size = (  
                        _direction
                        *min((_cap * self.max_cap_in_risk) / abs(signal.stop / _entry - 1), self.max_allowed_position_quoted) / _entry
                        / (len(ctx.instruments) if self.divide_by_symbols else 1)
                    )  
                    # fmt: on

                else:
                    logger.warning(
                        f" >>> {self.__class__.__name__}: stop is not specified for {str(signal)} - can't calculate position !"
                    )
                    continue

            t_pos.append(TargetPosition.create(ctx, signal, target_position_size))

        return t_pos


class WeightedPortfolioSizer(IPositionSizer):

    def __init__(self, capital_using: float = 1.0):
        """
        Weighted portfolio sizer. Signals are cosidered as weigths.

        :param cap_used_pct: how much of total capital to be used for sizing
        """
        self.capital_using = min(abs(capital_using), 1.0)

    def calculate_target_positions(self, ctx: StrategyContext, signals: List[Signal]) -> List[TargetPosition]:
        sw = np.sum([max(s.signal, 0) for s in signals])

        # - in case if all positions need to be closed
        sw = 1 if sw == 0 else sw

        cap = self.capital_using * ctx.get_capital()
        t_pos = []

        for signal in signals:
            # _pos = ctx.positions[signal.instrument.symbol]
            _q = ctx.quote(signal.instrument.symbol)
            if _q is not None:
                t_pos.append(
                    TargetPosition.create(
                        ctx,
                        signal,
                        cap * max(signal.signal, 0) / sw / _q.mid_price(),
                    )
                )
            else:
                logger.error(
                    f"{self.__class__.__name__}: Can't get actual market quote for {signal.instrument.symbol} !"
                )

        return t_pos



from typing import Any, Callable, Dict, List, Optional, Tuple, Union
from types import FunctionType


from threading import Thread
from multiprocessing.pool import ThreadPool
import traceback

import pandas as pd

from qubx import lookup, logger
from qubx.core.account import AccountProcessor
from qubx.core.helpers import BasicScheduler, CachedMarketDataHolder, process_schedule_spec, set_parameters_to_object
from qubx.core.loggers import LogsWriter, StrategyLogging
from qubx.core.basics import (
    TargetPosition,
    TriggerEvent,
    Deal,
    Instrument,
    Order,
    Position,
    Signal,
    dt_64,
    td_64,
    CtrlChannel,
    BatchEvent,
)
from qubx.core.loggers import StrategyLogging
from qubx.core.strategy import (
    IBrokerServiceProvider,
    IPositionGathering,
    IStrategy,
    ITradingServiceProvider,
    PositionsTracker,
    StrategyContext,
    SubscriptionType,
)
from qubx.core.series import Trade, Quote, Bar, OHLCV
from qubx.data.readers import DataReader
from qubx.gathering.simplest import SimplePositionGatherer
from qubx.trackers.sizers import FixedSizer
from qubx.utils.misc import Stopwatch
from qubx.utils.time import convert_seconds_to_str


_SW = Stopwatch()


def _dict_with_exception(dct, f):
    if f not in dct:
        raise ValueError(f"Configuration {dct} must contain field '{f}'")
    return dct[f]


class StrategyContextImpl(StrategyContext):
    """
    Main implementation of StrategyContext interface.
    """

    MAX_NUMBER_OF_STRATEGY_FAILURES = 10

    strategy: IStrategy  # strategy instance
    trading_service: ITradingServiceProvider  # service for exchange API: orders managemewnt
    broker_provider: IBrokerServiceProvider  # market data provider
    instruments: List[Instrument]  # list of instruments this strategy trades
    positions: Dict[str, Position]  # positions of the strategy (instrument -> position)
    acc: AccountProcessor
    positions_tracker: PositionsTracker  # position tracker
    positions_gathering: IPositionGathering  # position adjuster (executor)

    # - loggers
    _logging: StrategyLogging  # recording all activities for the strat: execs, positions, portfolio

    # - cached marked data anb scheduler
    _cache: CachedMarketDataHolder  # market data cache
    _scheduler: BasicScheduler

    # - aux data provider
    _aux_data_provider: DataReader | None  # auxiliary data provider

    # - configuration
    _market_data_subcription_type: str = "unknown"
    _market_data_subcription_params: dict = dict()
    _thread_data_loop: Thread | None = None  # market data loop

    _trig_interval_in_bar_nsec: int
    _trig_bar_freq_nsec: int
    _trig_on_bar: bool = False
    _trig_on_time: bool = False
    _trig_on_quote: bool = False
    _trig_on_trade: bool = False
    _trig_on_book: bool = False
    _current_bar_trigger_processed: bool = False
    _is_initilized: bool = False
    _symb_to_instr: Dict[str, Instrument]
    __strategy_id: str
    __order_id: int
    __fails_counter = 0
    __handlers: Dict[str, Callable[["StrategyContext", str, Any], TriggerEvent | None]]
    __fit_is_running: bool = False  # during fitting working it stops calling on_event method
    __init_fit_was_called: bool = False  # true if initial fit was already run
    __init_fit_args: Tuple = (None, None)  # arguments for initial on_fit() method call
    __pool: ThreadPool | None = None  # thread pool used for running aux tasks (fit etc)

    def __init__(
        self,
        # - strategy with parameters
        strategy: IStrategy,
        config: Dict[str, Any] | None,
        # - - - - - - - - - - - - - - - - - - - - -
        # - data provider and exchange service
        broker_connector: IBrokerServiceProvider,
        instruments: List[Instrument],
        base_currency: str = "USDT",
        initial_capital: float = 0,
        reserves: Dict[str, float] | None = None,
        # - - - - - - - - - - - - - - - - - - - - -
        # - data subscription - - - - - - - - - - -
        md_subscription: Dict[str, Any] = dict(type="ohlc", timeframe="1Min", nback=60),
        # - - - - - - - - - - - - - - - - - - - - -
        # - when need to trigger and fit strategy - - - - - - -
        trigger_spec: str | list[str] = "bar: -1Sec",  # 1 sec before subscription bar is closed
        fit_spec: str | None = None,
        # - - - - - - - - - - - - - - - - - - - - -
        # - how to write logs - - - - - - - - - -
        logs_writer: LogsWriter | None = None,
        positions_log_freq: str = "1Min",
        portfolio_log_freq: str = "5Min",
        num_exec_records_to_write=1,  # in live let's write every execution
        # - - - - - - - - - - - - - - - - - - - - -
        # - signals executor configuration - - - -
        position_gathering: IPositionGathering | None = None,
        # - - - - - - - - - - - - - - - - - - - - -
        # - aux data provider - - - - - - - - - - -
        aux_data_provider: DataReader | None = None,
    ) -> None:
        # - initialization
        self.broker_provider = broker_connector
        self.trading_service = broker_connector.get_trading_service()
        self.acc = AccountProcessor(
            account_id=self.trading_service.get_account_id(),
            base_currency=base_currency,
            reserves=reserves,
            initial_capital=initial_capital,
        )
        self.trading_service.set_account(self.acc)

        self.config = config
        self.instruments = instruments
        self.positions = {}
        self.strategy_name = strategy.__class__.__name__
        self.__fit_is_running = False
        self.__init_fit_was_called = False
        self.__pool = None
        self.__fails_counter = 0
        self._aux_data_provider = aux_data_provider

        # - for fast access to instrument by it's symbol
        self._symb_to_instr = {i.symbol: i for i in instruments}

        # - get scheduling service from broker
        self._scheduler = self.broker_provider.get_scheduler()

        # - instantiate logging functional
        self._logs_writer = logs_writer
        self._logging = StrategyLogging(logs_writer, positions_log_freq, portfolio_log_freq, num_exec_records_to_write)

        # - position adjuster
        self.positions_gathering = position_gathering if position_gathering else SimplePositionGatherer()

        # - extract data and event handlers
        self.__handlers = {
            n.split("_processing_")[1]: f
            for n, f in self.__class__.__dict__.items()
            if type(f) == FunctionType and n.startswith("_processing_")
        }

        # - create strategy instance and populate custom paramameters
        self._instantiate_strategy(strategy, config)

        # - process market data configuration
        self.__check_how_to_listen_to_market_data(md_subscription)

        # - process trigger and fit configurations
        self.__check_how_to_trigger_and_fit_strategy(trigger_spec, fit_spec)

        # - run cron scheduler
        self._scheduler.run()

    def _instantiate_strategy(self, strategy: IStrategy, config: Dict[str, Any] | None):
        # - set parameters to strategy (not sure we will do it here)
        self.strategy = strategy
        if isinstance(strategy, type):
            self.strategy = strategy()
        self.strategy.ctx = self

        # - set strategy custom parameters
        set_parameters_to_object(self.strategy, **config if config else {})
        self._is_initilized = False
        self.__strategy_id = self.strategy.__class__.__name__ + "_"
        self.__order_id = self.time().item() // 100_000_000

        # - here we need to get tracker
        _track = self.strategy.tracker(self)

        # - by default we use default position tracker with fixed 1:1 sizer
        #   so any signal is coinsidered as raw position size
        self.positions_tracker = _track if _track else PositionsTracker(FixedSizer(1.0, amount_in_quote=False))

    def __check_how_to_listen_to_market_data(self, md_config: dict):
        self._market_data_subcription_type = _dict_with_exception(md_config, "type").lower()
        match self._market_data_subcription_type:
            case "ohlc":
                timeframe = _dict_with_exception(md_config, "timeframe")
                self._market_data_subcription_params = {
                    "timeframe": timeframe,
                    "nback": md_config.get("nback", 1),
                }
                self._cache = CachedMarketDataHolder(timeframe)

            case "trade" | "trades" | "tas":
                timeframe = md_config.get("timeframe", "1Sec")
                self._market_data_subcription_params = {
                    "nback": md_config.get("nback", 1),
                }
                self._cache = CachedMarketDataHolder("1Sec")

            case "quote" | "quotes":
                self._cache = CachedMarketDataHolder("1Sec")

            case "ob" | "orderbook":
                self._cache = CachedMarketDataHolder("1Sec")

            case _:
                raise ValueError(
                    f"{self._market_data_subcription_type} is not a valid value for market data subcription type !!!"
                )

    def __check_how_to_trigger_and_fit_strategy(
        self, trigger_schedule: str | list[str] | None, fit_schedue: str | None
    ):
        _td2ns = lambda x: x.as_unit("ns").asm8.item()

        self._trig_interval_in_bar_nsec = 0

        if not trigger_schedule:
            raise ValueError("trigger parameter can't be empty !")
        if not isinstance(trigger_schedule, list):
            trigger_schedule = [trigger_schedule]

        t_rules = [process_schedule_spec(tr) for tr in trigger_schedule]

        f_rules = process_schedule_spec(fit_schedue)

        if not t_rules:
            raise ValueError(f"Couldn't recognize 'trigger' parameter specification: {trigger_schedule} !")

        # - we can use it as reference if bar timeframe is not secified
        _default_data_timeframe = pd.Timedelta(self._cache.default_timeframe)

        # - check trigger spec - - - - -
        for t_rule in t_rules:
            match t_rule["type"]:
                # it triggers on_event method when new bar is formed.
                # in this case it won't arm scheduler but just ruled by actual market data updates
                # - TODO: so probably need to drop this in favor to cron tasks
                # it's also possible to specify delay
                # "bar: -1s"      - it uses default timeframe and wake up 1 sec before every bar's close
                # "bar.5m: -5sec" - 5 sec before 5min bar's close
                # "bar.5m: 1sec"  - 1 sec after 5min bar closed (in next bar)
                case "bar":
                    _r_tf = t_rule.get("timeframe")
                    _bar_timeframe = _default_data_timeframe if not _r_tf else pd.Timedelta(_r_tf)
                    _inside_bar_delay: pd.Timedelta = t_rule.get("delay", pd.Timedelta(0))

                    if abs(pd.Timedelta(_inside_bar_delay)) > pd.Timedelta(_bar_timeframe):
                        raise ValueError(
                            f"Delay must be less or equal to bar's timeframe for bar trigger: you used {_inside_bar_delay} delay for {_bar_timeframe}"
                        )

                    # for positive delay - trigger strategy when this interval passed after new bar's open
                    if _inside_bar_delay >= pd.Timedelta(0):
                        self._trig_interval_in_bar_nsec = _td2ns(_inside_bar_delay)

                    # for negative delay - trigger strategy when time is closer to bar's closing time more than this interval
                    else:
                        self._trig_interval_in_bar_nsec = _td2ns(_bar_timeframe + _inside_bar_delay)

                    self._trig_bar_freq_nsec = _td2ns(_bar_timeframe)
                    self._trig_on_bar = True

                    logger.debug(
                        f"Triggering strategy on every {convert_seconds_to_str(self._trig_bar_freq_nsec/1e9)} bar after {convert_seconds_to_str(self._trig_interval_in_bar_nsec/1e9)}"
                    )

                case "cron":
                    if "schedule" not in t_rule:
                        raise ValueError(f"cron trigger type is specified but cron schedule not found !")

                    self._scheduler.schedule_event(t_rule["schedule"], "time_event")

                case "quote":
                    self._trig_on_quote = True
                    raise ValueError(f"quote trigger NOT IMPLEMENTED YET")

                case "trade":
                    self._trig_on_trade = True
                    logger.debug("Triggering strategy on every trade event")

                case "orderbook" | "ob":
                    self._trig_on_book = True
                    raise ValueError(f"orderbook trigger NOT IMPLEMENTED YET")

                case _:
                    raise ValueError(f"Wrong trigger type {t_rule['type']}")

        # - check fit spec - - - - -
        _last_fit_data_can_be_used = pd.Timestamp(self.time())
        match f_rules.get("type"):
            case "cron":
                if "schedule" not in f_rules:
                    raise ValueError(f"cron fit trigger type is specified but cron schedule not found !")

                self._scheduler.schedule_event(f_rules["schedule"], "fit_event")
                _last_fit_data_can_be_used = self._scheduler.get_event_last_time("fit_event")

            case "bar":
                raise ValueError("Raw 'bar' type is not supported for fitting spec yet, please use cron type !")

            case _:
                # if schedule is not specified just do not arm the task
                # only initial fit will be called
                pass

        # - we can't just call on_fit right now because not all market data may be ready
        # - so we just mark it as not called yet
        self.__init_fit_was_called = False
        self.__init_fit_args = (None, _last_fit_data_can_be_used)

    def process_data(self, symbol: str, d_type: str, data: Any) -> bool:
        # logger.debug(f" <cyan>({self.time()})</cyan> DATA : <yellow>{(symbol, d_type, data) }</yellow>")

        # - process data if handler is registered
        handler = self.__handlers.get(d_type)
        _SW.start("StrategyContext.handler")
        _strategy_trigger_on_event = handler(self, symbol, data) if handler else None
        _SW.stop("StrategyContext.handler")

        # - check if it still didn't call on_fit() for first time
        if not self.__init_fit_was_called:
            self._processing_fit_event(None, self.__init_fit_args)

        if _strategy_trigger_on_event:

            # - if fit was not called - skip on_event call
            if not self.__init_fit_was_called:
                logger.warning(
                    f"[{self.time()}] {self.strategy.__class__.__name__}::on_event() is SKIPPED for now because on_fit() was not called yet !"
                )
                return False

            # - if strategy still fitting - skip on_event call
            if self.__fit_is_running:
                logger.warning(
                    f"[{self.time()}] {self.strategy.__class__.__name__}::on_event() is SKIPPED for now because is being still fitting !"
                )
                return False

            signals: List[Signal] | Signal | None = None
            try:
                _SW.start("strategy.on_event")
                signals = self.strategy.on_event(self, _strategy_trigger_on_event)
                self._fails_counter = 0
            except Exception as strat_error:
                # - probably we need some cooldown interval after exception to prevent flooding
                logger.error(
                    f"[{self.time()}]: Strategy {self.strategy.__class__.__name__} raised an exception: {strat_error}"
                )
                logger.opt(colors=False).error(traceback.format_exc())

                #  - we stop execution after let's say maximal number of errors in a row
                self.__fails_counter += 1
                if self.__fails_counter >= StrategyContextImpl.MAX_NUMBER_OF_STRATEGY_FAILURES:
                    logger.error("STRATEGY FAILURES IN THE ROW EXCEEDED MAX ALLOWED NUMBER - STOPPING ...")
                    return True
            finally:
                _SW.stop("strategy.on_event")

            # - process and execute signals if they are provided
            if signals:
                # process signals by tracker and turn convert them into positions
                positions_from_strategy = self.__process_and_log_target_positions(
                    self.positions_tracker.process_signals(self, self.__process_signals(signals))
                )

                # gathering in charge of positions
                self.positions_gathering.alter_positions(self, positions_from_strategy)

        # - notify poition and portfolio loggers
        self._logging.notify(self.time())

        return False

    def _process_incoming_data_loop(self, channel: CtrlChannel):
        logger.info("(StrategyContext) Start processing market data")

        while channel.control.is_set():
            # - start loop latency measurement
            _SW.start("StrategyContext._process_incoming_data")

            # - waiting for incoming market data
            symbol, d_type, data = channel.receive()
            if self.process_data(symbol, d_type, data):
                _SW.stop("StrategyContext._process_incoming_data")
                channel.stop()
                break

            _SW.stop("StrategyContext._process_incoming_data")

        logger.info("(StrategyContext) Market data processing stopped")

    def _invoke_on_fit(self, current_fit_time: str | pd.Timestamp, prev_fit_time: str | pd.Timestamp | None):
        try:
            self.__fit_is_running = True
            logger.debug(
                f"Invoking <green>{self.strategy.__class__.__name__}</green> on_fit('{current_fit_time}', '{prev_fit_time}')"
            )
            _SW.start("strategy.on_fit")
            self.strategy.on_fit(self, current_fit_time, prev_fit_time)
            logger.debug(f"<green>{self.strategy.__class__.__name__}</green> is fitted")
        except Exception as strat_error:
            logger.error(
                f"[{self.time()}]: Strategy {self.strategy.__class__.__name__} on_fit('{current_fit_time}', '{prev_fit_time}') raised an exception: {strat_error}"
            )
            logger.opt(colors=False).error(traceback.format_exc())
        finally:
            self.__fit_is_running = False
            self.__init_fit_was_called = True
            _SW.stop("strategy.on_fit")
        return None

    def _processing_time_event(self, symbol: str, data: Any) -> TriggerEvent | None:
        """
        When scheduled time event is happened - we need to invoke strategy on_event method
        """
        return TriggerEvent(self.time(), "time", None, data)

    def _processing_fit_event(self, symbol: str | None, data: Any) -> TriggerEvent | None:
        """
        When scheduled fit event is happened - we need to invoke strategy on_fit method
        """
        if not self._cache.is_data_ready():
            return None

        # times are in seconds here
        prev_fit_time, now_fit_time = data

        # - we need to run this in separate thread
        self.__fit_is_running = True
        self._run_in_thread_pool(
            self._invoke_on_fit,
            (pd.Timestamp(now_fit_time, unit="s"), pd.Timestamp(prev_fit_time, unit="s") if prev_fit_time else None),
        )

        return None

    def _processing_hist_bar(self, symbol: str, bar: Bar) -> TriggerEvent | None:
        # - processing single historical bar
        #   here it just updates cache - historical bar can't trigger strategy logic
        self._cache.update_by_bar(symbol, bar)
        return None

    def _processing_hist_quote(self, symbol: str, quote: Quote | BatchEvent) -> TriggerEvent | None:
        if isinstance(quote, BatchEvent):
            for q in quote.data:
                self._cache.update_by_quote(symbol, q)
        else:
            self._cache.update_by_quote(symbol, quote)
        return None

    def _processing_hist_trade(self, symbol: str, trade: Trade | BatchEvent) -> TriggerEvent | None:
        if isinstance(trade, BatchEvent):
            for t in trade.data:
                self._cache.update_by_trade(symbol, t)
        else:
            self._cache.update_by_trade(symbol, trade)
        return None

    def _processing_hist_bars(self, symbol: str, bars: List[Bar]) -> TriggerEvent | None:
        # - processing historical bars as list
        for b in bars:
            self._processing_hist_bar(symbol, b)
        return None

    def _check_if_need_trigger_on_bar(self, symbol: str, bar: Bar | None):
        if self._trig_on_bar:
            t = self.trading_service.time().item()
            _time_to_trigger = t % self._trig_bar_freq_nsec >= self._trig_interval_in_bar_nsec
            if _time_to_trigger:
                # we want to trigger only first one - not every
                if not self._current_bar_trigger_processed:
                    self._current_bar_trigger_processed = True
                    if bar is None:
                        bar = self._cache.get_ohlcv(symbol)[0]
                    return TriggerEvent(self.time(), "bar", self._symb_to_instr.get(symbol), bar)
            else:
                self._current_bar_trigger_processed = False
        return None

    def __process_signals(self, signals: list[Signal] | Signal | None) -> List[Signal]:
        if isinstance(signals, Signal):
            signals = [signals]
        elif signals is None:
            return []

        for signal in signals:
            # set strategy group name if not set
            if not signal.group:
                signal.group = self.strategy_name

            # set reference prices for signals
            if signal.reference_price is None:
                q = self.quote(signal.instrument.symbol)
                if q is None:
                    continue
                signal.reference_price = q.mid_price()

        return signals

    def __process_signals_from_target_positions(
        self, target_positions: List[TargetPosition] | TargetPosition | None
    ) -> None:
        if target_positions is None:
            return
        if isinstance(target_positions, TargetPosition):
            target_positions = [target_positions]
        signals = [pos.signal for pos in target_positions]
        self.__process_signals(signals)

    def __process_and_log_target_positions(
        self, target_positions: List[TargetPosition] | TargetPosition | None
    ) -> List[TargetPosition]:

        if isinstance(target_positions, TargetPosition):
            target_positions = [target_positions]
        elif target_positions is None:
            return []

        self._logging.save_signals_targets(target_positions)
        return target_positions

    @_SW.watch("StrategyContext")
    def _processing_bar(self, symbol: str, bar: Bar) -> TriggerEvent | None:
        # - processing current bar's update
        self._cache.update_by_bar(symbol, bar)

        # - update tracker and handle alterd positions if need
        self.positions_gathering.alter_positions(
            self,
            self.__process_and_log_target_positions(
                self.positions_tracker.update(self, self._symb_to_instr[symbol], bar)
            ),
        )

        # - check if it's time to trigger the on_event if it's configured
        return self._check_if_need_trigger_on_bar(symbol, bar)

    def _processing_trade(self, symbol: str, trade: Trade | BatchEvent) -> TriggerEvent | None:
        is_batch_event = isinstance(trade, BatchEvent)
        if is_batch_event:
            for t in trade.data:
                self._cache.update_by_trade(symbol, t)
        else:
            self._cache.update_by_trade(symbol, trade)

        target_positions = self.positions_tracker.update(
            self, self._symb_to_instr[symbol], trade.data[-1] if is_batch_event else trade
        )
        self.__process_signals_from_target_positions(target_positions)

        # - update tracker and handle alterd positions if need
        self.positions_gathering.alter_positions(
            self,
            self.__process_and_log_target_positions(target_positions),
        )

        if self._trig_on_trade:
            event_type = "trade" if not is_batch_event else "batch:trade"
            return TriggerEvent(self.time(), event_type, self._symb_to_instr.get(symbol), trade)
        return None

    def _processing_quote(self, symbol: str, quote: Quote) -> TriggerEvent | None:
        self._cache.update_by_quote(symbol, quote)

        target_positions = self.positions_tracker.update(self, self._symb_to_instr[symbol], quote)
        self.__process_signals_from_target_positions(target_positions)

        # - update tracker and handle alterd positions if need
        self.positions_gathering.alter_positions(self, self.__process_and_log_target_positions(target_positions))

        # - TODO: here we can apply throttlings or filters
        #  - let's say we can skip quotes if bid & ask is not changed
        #  - or we can collect let's say N quotes before sending to strategy
        if self._trig_on_quote:
            return TriggerEvent(self.time(), "quote", self._symb_to_instr.get(symbol), quote)

        return self._check_if_need_trigger_on_bar(symbol, None)
        # return None

    @_SW.watch("StrategyContext")
    def _processing_order(self, symbol: str, order: Order) -> TriggerEvent | None:
        logger.debug(
            f"[<red>{order.id}</red> / {order.client_id}] : {order.type} {order.side} {order.quantity} of {symbol} { (' @ ' + str(order.price)) if order.price else '' } -> [{order.status}]"
        )
        # - check if we want to trigger any strat's logic on order
        return None

    def _processing_event(self, symbol: str, event_data: Dict) -> TriggerEvent | None:
        """
        Processing external events
        """
        return TriggerEvent(self.time(), "event", self._symb_to_instr.get(symbol), event_data)

    @_SW.watch("StrategyContext")
    def _processing_deals(self, symbol: str, deals: List[Deal]) -> TriggerEvent | None:
        # - log deals in storage
        instr = self._symb_to_instr.get(symbol)
        self._logging.save_deals(symbol, deals)
        if instr is not None:
            for d in deals:
                # - notify position gatherer and tracker
                self.positions_gathering.on_execution_report(self, instr, d)
                self.positions_tracker.on_execution_report(self, instr, d)
                logger.debug(f"Executed {d.amount} @ {d.price} of {symbol} for order <red>{d.order_id}</red>")
        else:
            logger.debug(f"Execution report for unknown instrument {symbol}")
        return None

    def ohlc(self, instrument: str | Instrument, timeframe: str) -> OHLCV:
        return self._cache.get_ohlcv(instrument if isinstance(instrument, str) else instrument.symbol, timeframe)

    def _create_and_update_positions(self, instruments: list[Instrument]):
        for instrument in instruments:
            symb = instrument.symbol
            self.positions[symb] = self.trading_service.get_position(instrument)

            # - check if we need any aux instrument for calculating pnl ?
            # TODO: test edge cases for aux symbols
            aux = lookup.find_aux_instrument_for(instrument, self.trading_service.get_base_currency())
            if aux is not None:
                instrument._aux_instrument = aux
                instruments.append(aux)
                self.positions[aux.symbol] = self.trading_service.get_position(aux)

    def start(self, blocking: bool = False):
        if self._is_initilized:
            raise ValueError("Strategy is already started !")

        # - create incoming market data processing
        databus = self.broker_provider.get_communication_channel()
        databus.register(self)

        self.__add_instruments(self.instruments)

        # - initialize strategy (should we do that after any first market data received ?)
        if not self._is_initilized:
            try:
                self.strategy.on_start(self)
                self._is_initilized = True
            except Exception as strat_error:
                logger.error(
                    f"(StrategyContext) Strategy {self.strategy.__class__.__name__} raised an exception in on_start: {strat_error}"
                )
                logger.error(traceback.format_exc())
                return

        # - for live we run loop
        if not self.broker_provider.is_simulated_trading:
            self._thread_data_loop = Thread(target=self._process_incoming_data_loop, args=(databus,), daemon=True)
            self._thread_data_loop.start()
            logger.info("(StrategyContext) strategy is started in thread")
            if blocking:
                self._thread_data_loop.join()

    def stop(self):
        if self._thread_data_loop:
            self.broker_provider.get_communication_channel().stop()
            self._thread_data_loop.join()
            try:
                self.strategy.on_stop(self)
            except Exception as strat_error:
                logger.error(
                    f"(StrategyContext) Strategy {self.strategy.__class__.__name__} raised an exception in on_stop: {strat_error}"
                )
                logger.error(traceback.format_exc())
            self._thread_data_loop = None

        # - close logging
        self._logging.close()
        self.get_latencies_report()

    def time(self) -> dt_64:
        return self.trading_service.time()

    def _generate_order_client_id(self, symbol: str) -> str:
        self.__order_id += 1
        return self.__strategy_id + symbol + "_" + str(self.__order_id)

    @_SW.watch("StrategyContext")
    def trade(
        self,
        instr_or_symbol: Instrument | str,
        amount: float,
        price: float | None = None,
        time_in_force="gtc",
        **options,
    ) -> Order:
        instrument: Instrument | None = (
            self._symb_to_instr.get(instr_or_symbol) if isinstance(instr_or_symbol, str) else instr_or_symbol
        )
        if instrument is None:
            raise ValueError(f"Can't find instrument for symbol {instr_or_symbol}")

        # - adjust size
        size_adj = instrument.round_size_down(abs(amount))
        if size_adj < instrument.min_size:
            raise ValueError(f"Attempt to trade size {abs(amount)} less than minimal allowed {instrument.min_size} !")

        side = "buy" if amount > 0 else "sell"
        type = "market"
        if price is not None:
            price = instrument.round_price_down(price) if amount > 0 else instrument.round_price_up(price)
            type = "limit"
            if (stp_type := options.get("stop_type")) is not None:
                type = f"stop_{stp_type}"

        logger.debug(
            f"(StrategyContext) sending {type} {side} for {size_adj} of <green>{instrument.symbol}</green> @ {price} ..."
        )
        client_id = self._generate_order_client_id(instrument.symbol)

        order = self.trading_service.send_order(
            instrument, side, type, size_adj, price, time_in_force=time_in_force, client_id=client_id, **options
        )

        return order

    @_SW.watch("StrategyContext")
    def cancel(self, instr_or_symbol: Instrument | str):
        instrument: Instrument | None = (
            self._symb_to_instr.get(instr_or_symbol) if isinstance(instr_or_symbol, str) else instr_or_symbol
        )
        if instrument is None:
            raise ValueError(f"Can't find instrument for symbol {instr_or_symbol}")
        for o in self.trading_service.get_orders(instrument.symbol):
            self.trading_service.cancel_order(o.id)

    def cancel_order(self, order_id: str):
        if order_id:
            self.trading_service.cancel_order(order_id)

    def get_capital(self) -> float:
        return self.trading_service.get_capital()

    def get_reserved(self, instrument: Instrument) -> float:
        return self.trading_service.get_account().get_reserved(instrument)

    @_SW.watch("StrategyContext")
    def set_universe(self, instruments: list[Instrument]) -> None:
        for instr in instruments:
            self._symb_to_instr[instr.symbol] = instr

        new_set = set(instruments)
        prev_set = set(self.instruments)
        rm_instr = list(prev_set - new_set)
        add_instr = list(new_set - prev_set)

        self.__add_instruments(add_instr)
        self.__remove_instruments(rm_instr)

        if add_instr or rm_instr:
            self.strategy.on_universe_change(self, add_instr, rm_instr)

        # set new instruments
        self.instruments = instruments

    def __remove_instruments(self, instruments: list[Instrument]) -> None:
        """
        Remove symbols from universe. Steps:
        - close all open positions
        - unsubscribe from market data
        - remove from data cache

        We are still keeping the symbols in the positions dictionary.
        """
        # - close all open positions
        exit_targets = [
            TargetPosition.zero(self, instr.signal(0, group="Universe", comment="Universe change"))
            for instr in instruments
            if instr.symbol in self.positions and abs(self.positions[instr.symbol].quantity) > instr.min_size
        ]
        self.positions_gathering.alter_positions(self, exit_targets)

        # - if still open positions close them manually
        for instr in instruments:
            pos = self.positions.get(instr.symbol)
            if pos and abs(pos.quantity) > instr.min_size:
                self.trade(instr, -pos.quantity)

        # - unsubscribe from market data
        for instr in instruments:
            self.unsubscribe(self._market_data_subcription_type, instr)

        # - remove from data cache
        for instr in instruments:
            self._cache.remove(instr.symbol)

    def __add_instruments(self, instruments: list[Instrument]) -> None:
        # - create positions for instruments
        self._create_and_update_positions(instruments)

        # - get actual positions from exchange
        _symbols = []
        for instr in instruments:
            # process instruments - need to find convertors etc
            self._cache.init_ohlcv(instr.symbol)
            _symbols.append(instr.symbol)

        # - subscribe to market data
        logger.debug(
            f"(StrategyContext) Subscribing to {self._market_data_subcription_type} updates using {self._market_data_subcription_params} for \n\t{_symbols} "
        )
        self.broker_provider.subscribe(
            self._market_data_subcription_type, instruments, **self._market_data_subcription_params
        )

        # - initialize strategy loggers
        self._logging.initialize(self.time(), self.positions, self.trading_service.get_account().get_balances())

    @_SW.watch("StrategyContext")
    def subscribe(self, subscription_type: str, instr_or_symbol: Instrument | str, **kwargs) -> bool:
        """
        Subscribe to market data updates
        """
        instrument: Instrument | None = (
            self._symb_to_instr.get(instr_or_symbol) if isinstance(instr_or_symbol, str) else instr_or_symbol
        )
        if instrument is None:
            raise ValueError(f"Can't find instrument for symbol {instr_or_symbol}")

        return self.broker_provider.subscribe(subscription_type, [instrument], **kwargs)

    @_SW.watch("StrategyContext")
    def unsubscribe(self, subscription_type: str, instr_or_symbol: Instrument | str) -> bool:
        """
        Unsubscribe from market data updates
        """
        instrument: Instrument | None = (
            self._symb_to_instr.get(instr_or_symbol) if isinstance(instr_or_symbol, str) else instr_or_symbol
        )
        if instrument is None:
            raise ValueError(f"Can't find instrument for symbol {instr_or_symbol}")

        return self.broker_provider.unsubscribe(subscription_type, [instrument])

    def has_subscription(self, subscription_type: str, instr_or_symbol: Instrument | str) -> bool:
        """
        Check if subscription is active
        """
        instrument: Instrument | None = (
            self._symb_to_instr.get(instr_or_symbol) if isinstance(instr_or_symbol, str) else instr_or_symbol
        )
        if instrument is None:
            raise ValueError(f"Can't find instrument for symbol {instr_or_symbol}")

        return self.broker_provider.has_subscription(subscription_type, instrument)

    def _run_in_thread_pool(self, func: Callable, args=()):
        """
        For the simulation we don't need to call function in thread
        """
        if self.broker_provider.is_simulated_trading:
            func(*args)
        else:
            if self.__pool is None:
                self.__pool = ThreadPool(2)
            self.__pool.apply_async(func, args)

    # - IMarketDataProvider methods implementation -

    @_SW.watch("StrategyContext")
    def get_historical_ohlcs(self, instrument: Instrument | str, timeframe: str, length: int) -> OHLCV | None:
        """
        Helper for historical ohlc data
        """
        instr = self._symb_to_instr.get(instrument) if isinstance(instrument, str) else instrument

        if instr is None:
            logger.warning(f"Can't find instrument for {instrument} symbol !")
            return None

        # - first check if we can use cached series
        rc = self.ohlc(instr, timeframe)
        if len(rc) >= length:
            return rc

        # - send request for historical data
        bars = self.broker_provider.get_historical_ohlcs(instr.symbol, timeframe, length)
        r = self._cache.update_by_bars(instr.symbol, timeframe, bars)
        return r

    def get_aux_data(self, data_id: str, **parameters) -> pd.DataFrame | None:
        if self._aux_data_provider:
            return self._aux_data_provider.get_aux_data(data_id, **parameters)
        return None

    def get_instruments(self) -> List[Instrument]:
        return self.instruments

    def quote(self, symbol: str) -> Quote | None:
        return self.broker_provider.get_quote(symbol)

    def get_instrument(self, symbol: str) -> Instrument | None:
        return self._symb_to_instr.get(symbol)
